@@ -1,6 +1,6 @@
 
 /*
- * Copyright (c) 2007 - 2009 Vreixo Formoso, Thomas Schmitt
+ * Copyright (c) 2007 - 2015 Vreixo Formoso, Thomas Schmitt
  * 
  * This file is part of the libisofs project; you can redistribute it and/or 
  * modify it under the terms of the GNU General Public License version 2 
@@ -24,7 +24,9 @@ static char helptext[][80] = {
 "               Output the contents of an iso image.",
 "  -iso_cat  image_file path_in_image",
 "               Extract a file from a given ISO image and put out its content",
-"               to stdout. The file is addressed by path_in_image.",
+"               to stdout. The file is addressed by path_in_image. The ISO",
+"               image does not get loaded but rather the lookups are done",
+"               directly in the image file.",
 "  -iso_modify  image_file absolute_directory_path output_file",
 "               Load an iso image, add a directory, and write complete image.",
 "  -iso_ms  image_lba nwa image_file directory_path output_file",
@@ -58,6 +60,22 @@ static char helptext[][80] = {
 #endif
 
 
+/* ----------------------------- utilities -------------------------- */
+
+
+void demo_report_iso_err(int err, char *occasion)
+{
+    char *severity;
+
+    fprintf(stderr, "%s : err = 0x%X", occasion, (unsigned int) err);
+    if (err < 0) {
+        iso_sev_to_text(iso_error_get_severity(err), &severity);
+        fprintf(stderr, " -> %s '%s'", severity, iso_error_to_msg(err));
+    }
+    fprintf(stderr, "\n");
+}
+
+
 /* ------------------------- from demo/tree.c ----------------------- */
 
 static void
@@ -446,7 +464,8 @@ iso_read_print_dir(IsoFileSource *dir, int level)
         sp[i+1] = ' ';
     }
 
-    sp[level * 2-1] = '-';
+    if (level > 0)
+        sp[level * 2 - 1] = '-';
     sp[level * 2] = '\0';
 
     ret = iso_file_source_open(dir);
@@ -474,37 +493,43 @@ iso_read_print_dir(IsoFileSource *dir, int level)
 
 int gesture_iso_read(int argc, char **argv)
 {
-    int result;
-    IsoImageFilesystem *fs;
-    IsoDataSource *src;
-    IsoFileSource *root;
-    IsoReadOpts *ropts;
+    int result, initialized = 0, return_val = 1;
+    IsoImageFilesystem *fs = NULL;
+    IsoDataSource *src = NULL;
+    IsoFileSource *root = NULL;
+    IsoReadOpts *ropts = NULL;
 
     if (argc != 2) {
         printf ("You need to specify a valid path\n");
-        return 1;
+        goto ex;
     }
 
-    iso_init();
+    result = iso_init();
+    if (result < 0) {
+        demo_report_iso_err(result, "Cannot init libisofs");
+        goto ex;
+    }
+    initialized = 1;
+
     iso_set_msgs_severities("NEVER", "ALL", "");
 
     result = iso_data_source_new_from_file(argv[1], &src);
     if (result < 0) {
-        printf ("Error creating data source\n");
-        return 1;
+        demo_report_iso_err(result, "Error creating data source");
+        goto ex;
     }
-
     result = iso_read_opts_new(&ropts, 0);
     if (result < 0) {
-        fprintf(stderr, "Error creating read options\n");
-        return 1;
+        demo_report_iso_err(result, "Error creating read options");
+        goto ex;
     }
     result = iso_image_filesystem_new(src, ropts, 1, &fs);
-    iso_read_opts_free(ropts);
     if (result < 0) {
-        printf ("Error creating filesystem\n");
-        return 1;
+        demo_report_iso_err(result, "Error creating filesystem");
+        goto ex;
     }
+    iso_read_opts_free(ropts);
+    ropts = NULL;
 
     printf("\nVOLUME INFORMATION\n");
     printf("==================\n\n");
@@ -523,18 +548,27 @@ int gesture_iso_read(int argc, char **argv)
 
     result = fs->get_root(fs, &root);
     if (result < 0) {
-        printf ("Can't get root %d\n", result);
-        return 1;
+        demo_report_iso_err(result, "Cannot get root object");
+        goto ex;
     }
     /* iso_read_print_file_src(root); */
     iso_read_print_dir(root, 0);
-    iso_file_source_unref(root);
 
-    fs->close(fs);
-    iso_filesystem_unref((IsoFilesystem*)fs);
-    iso_data_source_unref(src);
-    iso_finish();
-    return 0;
+    return_val = 0;
+ex:;
+    if (root != NULL)
+        iso_file_source_unref(root);
+    if (ropts != NULL)
+        iso_read_opts_free(ropts);
+    if (fs != NULL) {
+        fs->close(fs);
+        iso_filesystem_unref((IsoFilesystem*)fs);
+    }
+    if (src != NULL)
+        iso_data_source_unref(src);
+    if (initialized)
+        iso_finish();
+    return return_val;
 }
 
 
@@ -543,83 +577,101 @@ int gesture_iso_read(int argc, char **argv)
 
 int gesture_iso_cat(int argc, char **argv)
 {
-    int res, write_ret;
-    IsoFilesystem *fs;
-    IsoFileSource *file;
+    int res, write_ret, ret;
+    IsoFilesystem *fs = NULL;
+    IsoFileSource *file = NULL;
     struct stat info;
-    IsoDataSource *src;
-    IsoReadOpts *opts;
+    IsoDataSource *src = NULL;
+    IsoReadOpts *opts = NULL;
 
     if (argc != 3) {
-        fprintf(stderr, "Usage: isocat /path/to/image /path/to/file\n");
+        fprintf(stderr, "Usage: -iso_cat /path/to/image /path/to/file\n");
         return 1;
     }
 
     res = iso_init();
     if (res < 0) {
-        fprintf(stderr, "Can't init libisofs\n");
+        demo_report_iso_err(res, "Cannot init libisofs");
         return 1;
     }
-    
+
+    /* Important Note:
+       From here on memory objects get created which need to be freed in
+       the end. Therefore in case of problems no direct return, but rather
+       a hop to label "ex:", where cleanup happens.
+     */
+
     res = iso_data_source_new_from_file(argv[1], &src);
     if (res < 0) {
-        fprintf(stderr, "Error creating data source\n");
-        return 1;
+        demo_report_iso_err(res, "Error creating data source object");
+        ret = 1; goto ex;
     }
 
     res = iso_read_opts_new(&opts, 0);
     if (res < 0) {
-        fprintf(stderr, "Error creating read options\n");
-        return 1;
+        demo_report_iso_err(res, "Error creating read options object");
+        ret = 1; goto ex;
     }
     res = iso_image_filesystem_new(src, opts, 1, &fs);
     if (res < 0) {
-        fprintf(stderr, "Error creating filesystem\n");
-        return 1;
+        demo_report_iso_err(res, "Error creating filesystem object");
+        ret = 1; goto ex;
     }
     iso_read_opts_free(opts);
+    opts = NULL;
 
     res = fs->get_by_path(fs, argv[2], &file);
     if (res < 0) {
-        fprintf(stderr, "Can't get file, err = %d\n", res);
-        return 1;
+        demo_report_iso_err(res, "Cannot get file object with given path");
+        ret = 1; goto ex;
     }
 
     res = iso_file_source_lstat(file, &info);
     if (res < 0) {
-        fprintf(stderr, "Can't stat file, err = %d\n", res);
-        return 1;
+        demo_report_iso_err(res,
+                         "Cannot inquire type of file object with given path");
+        ret = 1; goto ex;
     }
 
     if (S_ISDIR(info.st_mode)) {
         fprintf(stderr, "Path refers to a directory!!\n");
-        return 1;
+        ret = 1; goto ex;
     } else {
         char buf[1024];
         res = iso_file_source_open(file);
         if (res < 0) {
-            fprintf(stderr, "Can't open file, err = %d\n", res);
-            return 1;
+            demo_report_iso_err(res,
+                                "Cannot open file object with given path");
+            ret = 1; goto ex;
         }
         while ((res = iso_file_source_read(file, buf, 1024)) > 0) {
             write_ret = fwrite(buf, 1, res, stdout);
             if (write_ret < res) {
                 printf ("Cannot write block to stdout. errno= %d\n", errno);
-                return 1;
+                iso_file_source_close(file);
+                ret = 1; goto ex;
             }
         }
-        if (res < 0) {
-            fprintf(stderr, "Error reading, err = %d\n", res);
-            return 1;
-        }
         iso_file_source_close(file);
+        if (res < 0) {
+            demo_report_iso_err(res, "Error while reading data content");
+            fprintf(stderr, "Error reading, err = 0x%X\n", (unsigned int) res);
+            ret = 1; goto ex;
+        }
     }
-    
-    iso_file_source_unref(file);
-    iso_filesystem_unref(fs);
-    iso_data_source_unref(src);
+
+    ret = 0;
+ex:;
+    if (file != NULL)
+        iso_file_source_unref(file);
+    if (fs != NULL)
+        iso_filesystem_unref(fs);
+    if (opts != NULL)
+        iso_read_opts_free(opts);
+    if (src != NULL)
+        iso_data_source_unref(src);
     iso_finish();
-    return 0;
+    return ret;
 }
 
 
@@ -633,14 +685,14 @@ void iso_modify_usage(char **argv)
 
 int gesture_iso_modify(int argc, char **argv)
 {
-    int result;
-    IsoImage *image;
-    IsoDataSource *src;
-    struct burn_source *burn_src;
+    int result, return_val = 1, initialized = 0;
+    IsoImage *image = NULL;
+    IsoDataSource *src = NULL;
+    struct burn_source *burn_src = NULL;
     unsigned char buf[2048];
     FILE *fp = NULL;
-    IsoWriteOpts *opts;
-    IsoReadOpts *ropts;
+    IsoWriteOpts *opts = NULL;
+    IsoReadOpts *ropts = NULL;
 	
     if (argc < 4) {
         iso_modify_usage(argv);
@@ -653,20 +705,25 @@ int gesture_iso_modify(int argc, char **argv)
         goto ex;
     }
 
-    iso_init();
+    result = iso_init();
+    if (result < 0) {
+        demo_report_iso_err(result, "Cannot init libisofs");
+        goto ex;
+    }
+    initialized = 1;
     iso_set_msgs_severities("NEVER", "ALL", "");
     
     /* create the data source to accesss previous image */
     result = iso_data_source_new_from_file(argv[1], &src);
     if (result < 0) {
-        printf ("Error creating data source\n");
+        demo_report_iso_err(result, "Error creating data source");
         goto ex;
     }
     
     /* create the image context */
     result = iso_image_new("volume_id", &image);
     if (result < 0) {
-        printf ("Error creating image\n");
+        demo_report_iso_err(result, "Error creating image");
         goto ex;
     }
     iso_tree_set_follow_symlinks(image, 0);
@@ -675,58 +732,75 @@ int gesture_iso_modify(int argc, char **argv)
     /* import previous image */
     result = iso_read_opts_new(&ropts, 0);
     if (result < 0) {
-        fprintf(stderr, "Error creating read options\n");
+        demo_report_iso_err(result, "Error creating read options");
         goto ex;
     }
     result = iso_image_import(image, src, ropts, NULL);
-    iso_read_opts_free(ropts);
-    iso_data_source_unref(src);
     if (result < 0) {
-        printf ("Error importing previous session %d\n", result);
+        demo_report_iso_err(result, "Error importing previous session");
         goto ex;
     }
+    /* (One could of course keep them alive until cleanup) */
+    iso_read_opts_free(ropts);
+    ropts = NULL;
+    iso_data_source_unref(src);
+    src = NULL;
     
     /* add new dir */
     result = iso_tree_add_dir_rec(image, iso_image_get_root(image), argv[2]);
     if (result < 0) {
-        printf ("Error adding directory %d\n", result);
+        demo_report_iso_err(result, "Error adding directory");
         goto ex;
     }
     
-    /* generate a new image with both previous and added contents */
+    /* Generate a new image with both previous and added contents.
+       Profile 1 means Rock Ridge and ISO level 3.
+    */
     result = iso_write_opts_new(&opts, 1);
     if (result < 0) {
-        printf("Cant create write opts, error %d\n", result);
+        demo_report_iso_err(result, "Cannot create write opts");
         goto ex;
     }
-    /* for isolinux: iso_write_opts_set_allow_full_ascii(opts, 1); */
-    
+    /* Prefer specs violation over relocation deep directories */
+    iso_write_opts_set_allow_deep_paths(opts, 1);
+
+    /* For MS-Windows readers : iso_write_opts_set_joliet(opts, 1); */
+ 
     result = iso_image_create_burn_source(image, opts, &burn_src);
     if (result < 0) {
-        printf ("Cant create image, error %d\n", result);
+        demo_report_iso_err(result, "Cannot create image object");
         goto ex;
     }
-    
     iso_write_opts_free(opts);
+    opts = NULL;
     
     while (burn_src->read_xt(burn_src, buf, 2048) == 2048) {
         result = fwrite(buf, 1, 2048, fp);
         if (result < 2048) {
-            printf ("Cannot write block. errno= %d\n", errno);
+            fprintf (stderr, "Cannot write block. errno= %d\n", errno);
             goto ex;
         }
     }
-    fclose(fp);
-    burn_src->free_data(burn_src);
-    free(burn_src);
-    
-    iso_image_unref(image);
-    iso_finish();
-    return 0;
+
+    return_val = 0;
 ex:
     if (fp != NULL)
         fclose(fp);
-    return 1;
+    if (opts != NULL)
+        iso_write_opts_free(opts);
+    if (burn_src != NULL) {
+        burn_src->free_data(burn_src);
+        free(burn_src);
+    }
+    if (image != NULL)
+        iso_image_unref(image);
+    if (ropts != NULL)
+        iso_read_opts_free(ropts);
+    if (src != NULL)
+        iso_data_source_unref(src);
+    if (initialized)
+        iso_finish();
+    return return_val;
 }
 
 
@@ -740,14 +814,14 @@ void iso_ms_usage(char **argv)
 
 int gesture_iso_ms(int argc, char **argv)
 {
-    int result;
-    IsoImage *image;
-    IsoDataSource *src;
-    struct burn_source *burn_src;
+    int result, return_val = 1, initialized = 0;
+    IsoImage *image = NULL;
+    IsoDataSource *src = NULL;
+    struct burn_source *burn_src = NULL;
     unsigned char buf[2048];
     FILE *fp = NULL;
-    IsoWriteOpts *opts;
-    IsoReadOpts *ropts;
+    IsoWriteOpts *opts = NULL;
+    IsoReadOpts *ropts = NULL;
     uint32_t ms_block;
 	
     if (argc < 6) {
@@ -767,20 +841,26 @@ int gesture_iso_ms(int argc, char **argv)
         goto ex;
     }
 
-    iso_init();
+    result = iso_init();
+    if (result < 0) {
+        demo_report_iso_err(result, "Cannot init libisofs");
+        goto ex;
+    }
+    initialized = 1;
+
     iso_set_msgs_severities("NEVER", "ALL", "");
     
     /* create the data source to accesss previous image */
     result = iso_data_source_new_from_file(argv[3], &src);
     if (result < 0) {
-        printf ("Error creating data source\n");
+        demo_report_iso_err(result, "Error creating data source");
         goto ex;
     }
     
     /* create the image context */
     result = iso_image_new("volume_id", &image);
     if (result < 0) {
-        printf ("Error creating image\n");
+        demo_report_iso_err(result, "Error creating image");
         goto ex;
     }
     iso_tree_set_follow_symlinks(image, 0);
@@ -795,23 +875,25 @@ int gesture_iso_ms(int argc, char **argv)
     iso_read_opts_set_start_block(ropts, atoi(argv[1]));
     result = iso_image_import(image, src, ropts, NULL);
     iso_read_opts_free(ropts);
+    ropts = NULL;
     iso_data_source_unref(src);
+    src = NULL;
     if (result < 0) {
-        printf ("Error importing previous session %d\n", result);
+        demo_report_iso_err(result, "Error importing previous session");
         goto ex;
     }
     
     /* add new dir */
     result = iso_tree_add_dir_rec(image, iso_image_get_root(image), argv[4]);
     if (result < 0) {
-        printf ("Error adding directory %d\n", result);
+        demo_report_iso_err(result, "Error adding directory");
         goto ex;
     }
     
     /* generate a multisession image with new contents */
     result = iso_write_opts_new(&opts, 1);
     if (result < 0) {
-        printf("Cant create write opts, error %d\n", result);
+        demo_report_iso_err(result, "Cannot create write opts");
         goto ex;
     }
     
@@ -826,6 +908,7 @@ int gesture_iso_ms(int argc, char **argv)
         goto ex;
     }
     iso_write_opts_free(opts);
+    opts = NULL;
     
     while (burn_src->read_xt(burn_src, buf, 2048) == 2048) {
         result = fwrite(buf, 1, 2048, fp);
@@ -834,17 +917,26 @@ int gesture_iso_ms(int argc, char **argv)
             goto ex;
         }
     }
-    fclose(fp);
-    burn_src->free_data(burn_src);
-    free(burn_src);
-    
-    iso_image_unref(image);
-    iso_finish();
-    return 0;
+
+    return_val = 0;
 ex:;
+    if (burn_src != NULL) {
+        burn_src->free_data(burn_src);
+        free(burn_src);
+    }
+    if (opts != NULL)
+        iso_write_opts_free(opts);
+    if (image)
+        iso_image_unref(image);
+    if (ropts != NULL)
+        iso_read_opts_free(ropts);
+    if (src != NULL)
+        iso_data_source_unref(src);
+    if (initialized)
+        iso_finish();
     if (fp != NULL)
         fclose(fp);
-    return 1;
+    return return_val;
 }
 
 
