@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
- * Copyright (c) 2009 - 2011 Thomas Schmitt
+ * Copyright (c) 2009 - 2015 Thomas Schmitt
  * 
  * This file is part of the libisofs project; you can redistribute it and/or 
  * modify it under the terms of the GNU General Public License version 2 
@@ -21,6 +21,7 @@
 #include "image.h"
 #include "aaip_0_2.h"
 #include "util.h"
+#include "messages.h"
 
 #include <stdlib.h>
 #include <string.h>
@@ -72,8 +73,15 @@ int default_create_file(IsoNodeBuilder *builder, IsoImage *image,
     iso_file_source_ref(src);
     
     name = iso_file_source_get_name(src);
-    if (strlen(name) > LIBISOFS_NODE_NAME_MAX)
-        name[LIBISOFS_NODE_NAME_MAX] = 0;
+    if ((int) strlen(name) > image->truncate_length) {
+        ret = iso_truncate_rr_name(image->truncate_mode,
+                                   image->truncate_length, name, 0);
+        if (ret < 0) {
+            iso_stream_unref(stream);
+            free(name);
+            return ret;
+        }
+    }
     ret = iso_node_new_file(name, stream, &node);
     if (ret < 0) {
         iso_stream_unref(stream);
@@ -96,13 +104,13 @@ int default_create_file(IsoNodeBuilder *builder, IsoImage *image,
 
 static
 int default_create_node(IsoNodeBuilder *builder, IsoImage *image,
-                        IsoFileSource *src, IsoNode **node)
+                        IsoFileSource *src, char *in_name, IsoNode **node)
 {
-    int ret;
+    int ret, name_is_attached = 0;
     struct stat info;
     IsoNode *new;
     IsoFilesystem *fs;
-    char *name;
+    char *name = NULL;
     unsigned char *aa_string = NULL;
     char *a_text = NULL, *d_text = NULL;
     char *dest = NULL;
@@ -122,9 +130,21 @@ int default_create_node(IsoNodeBuilder *builder, IsoImage *image,
         goto ex;
     }
 
-    name = iso_file_source_get_name(src);
-    if (strlen(name) > LIBISOFS_NODE_NAME_MAX)
-        name[LIBISOFS_NODE_NAME_MAX] = 0;
+    if (in_name == NULL) {
+        name = iso_file_source_get_name(src);
+    } else {
+        name = strdup(in_name);
+        if (name == NULL) {
+            ret = ISO_OUT_OF_MEM; goto ex;
+        }
+    }
+
+    if ((int) strlen(name) > image->truncate_length) {
+        ret = iso_truncate_rr_name(image->truncate_mode,
+                                   image->truncate_length, name, 0);
+        if (ret < 0)
+            goto ex;
+    }
     fs = iso_file_source_get_filesystem(src);
     new = NULL;
 
@@ -195,12 +215,14 @@ int default_create_node(IsoNodeBuilder *builder, IsoImage *image,
             }
         }
         break;
-    }
-    
-    if (ret < 0) {
-        free(name);
+    default:
+        ret = ISO_BAD_FSRC_FILETYPE;
         goto ex;
     }
+    
+    if (ret < 0)
+        goto ex;
+    name_is_attached = 1;
 
     /* fill fields */
     iso_node_set_perms_internal(new, info.st_mode, 1);
@@ -214,9 +236,9 @@ int default_create_node(IsoNodeBuilder *builder, IsoImage *image,
     /* Eventually set S_IRWXG from ACL */
     if (image->builder_ignore_acl) {
         ret = iso_file_source_get_aa_string(src, &aa_string, 4);
-        if (aa_string != NULL)
+        if (ret >= 0 && aa_string != NULL)
             iso_aa_get_acl_text(aa_string, info.st_mode, &a_text, &d_text, 16);
-        if (a_text != NULL) {
+        if (ret >= 0 && a_text != NULL) {
             aaip_cleanout_st_mode(a_text, &(info.st_mode), 4 | 16);
             iso_node_set_perms_internal(new, info.st_mode, 1);
         }
@@ -243,6 +265,8 @@ int default_create_node(IsoNodeBuilder *builder, IsoImage *image,
 
     ret = ISO_SUCCESS;
 ex:;
+    if (name != NULL && !name_is_attached)
+        free(name);
     LIBISO_FREE_MEM(dest);
     return ret;
 }
