@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
- * Copyright (c) 2010 - 2013 Thomas Schmitt
+ * Copyright (c) 2010 - 2014 Thomas Schmitt
  *
  * This file is part of the libisofs project; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License version 2 
@@ -82,15 +82,16 @@ void el_torito_set_load_seg(ElToritoBootImage *bootimg, short segment)
 {
     if (bootimg->type != 0)
         return;
-    bootimg->load_seg = segment;
+    if (segment < 0)
+        bootimg->load_seg = 0x1000 + segment;
+    else
+        bootimg->load_seg = segment;
 }
 
 /* API */
 int el_torito_get_load_seg(ElToritoBootImage *bootimg)
 {
-   if (bootimg->load_seg < 0)
-       return 0xffff - bootimg->load_seg;
-   return bootimg->load_seg;
+   return (int) bootimg->load_seg;
 }
  
 /**
@@ -102,15 +103,16 @@ void el_torito_set_load_size(ElToritoBootImage *bootimg, short sectors)
 {
     if (bootimg->type != 0)
         return;
-    bootimg->load_size = sectors;
+    if (sectors < 0)
+        bootimg->load_size = 0x10000 + sectors;
+    else
+        bootimg->load_size = sectors;
 }
 
 /* API */
 int el_torito_get_load_size(ElToritoBootImage *bootimg)
 {
-   if (bootimg->load_size < 0)
-       return 0xffff - bootimg->load_size;
-   return bootimg->load_size;
+   return (int) bootimg->load_size;
 }
 
 /**
@@ -204,6 +206,8 @@ void el_torito_patch_isolinux_image(ElToritoBootImage *bootimg)
 int el_torito_set_isolinux_options(ElToritoBootImage *bootimg, int options, int flag)
 {
     bootimg->isolinux_options = (options & 0x03ff);
+    bootimg->seems_boot_info_table = !!(options & 1);
+    bootimg->seems_grub2_boot_info = !!(options & (1 << 9));
     return ISO_SUCCESS;
 }
 
@@ -440,6 +444,8 @@ int create_image(IsoImage *image, const char *image_path,
     iso_node_ref(imgfile); /* get our ref */
     boot->bootable = 1;
     boot->seems_boot_info_table = 0;
+    boot->seems_grub2_boot_info = 0;
+    boot->seems_isohybrid_capable = 0;
     boot->isolinux_options = 0;
     boot->type = boot_media_type;
     boot->partition_type = partition_type;
@@ -448,9 +454,7 @@ int create_image(IsoImage *image, const char *image_path,
     boot->platform_id = 0; /* 80x86 */
     memset(boot->id_string, 0, sizeof(boot->id_string));
     memset(boot->selection_crit, 0, sizeof(boot->selection_crit));
-    if (bootimg) {
-        *bootimg = boot;
-    }
+    *bootimg = boot;
 
     return ISO_SUCCESS;
 }
@@ -533,8 +537,8 @@ int iso_image_set_boot_image(IsoImage *image, const char *image_path,
     for (i = 1; i < Libisofs_max_boot_imageS; i++)
         catalog->bootimages[i] = NULL;
     catalog->node = cat_node;
-    catalog->sort_weight = 1000;                            /* slightly high */
-    if (!boot_node->explicit_weight)
+    catalog->sort_weight = 1000000000;                          /* very high */
+    if (!(boot_node->explicit_weight || boot_node->from_old_session))
         boot_node->sort_weight = 2;
     iso_node_ref((IsoNode*)cat_node);
     image->bootcat = catalog;
@@ -715,7 +719,7 @@ int iso_image_add_boot_image(IsoImage *image, const char *image_path,
     ret = create_image(image, image_path, type, &boot_img, &boot_node);
     if (ret < 0) 
         return ret;
-    if (!boot_node->explicit_weight)
+    if (!(boot_node->explicit_weight || boot_node->from_old_session))
         boot_node->sort_weight = 2;
     catalog->bootimages[catalog->num_bootimages] = boot_img;
     catalog->num_bootimages++;
@@ -808,7 +812,8 @@ write_section_header(uint8_t *buf, Ecma119Image *t, int idx, int num_entries)
         (struct el_torito_section_header *) buf;
 
     /* 0x90 = more section headers follow , 0x91 = final section */
-    e->header_indicator[0] = 0x90 + (idx == t->catalog->num_bootimages - 1);
+    e->header_indicator[0] =
+                      0x90 + (idx == t->catalog->num_bootimages - num_entries);
     e->platform_id[0] = t->catalog->bootimages[idx]->platform_id;
     e->num_entries[0] = num_entries & 0xff;
     e->num_entries[1] = (num_entries >> 8) & 0xff;;
@@ -951,7 +956,8 @@ int catalog_is_repeatable(IsoStream *stream)
 /**
  * fs_id will be the id reserved for El-Torito
  * dev_id will be 0 for catalog, 1 for boot image (if needed)
- * we leave ino_id for future use when we support multiple boot images
+ * ino_id 0 is supposed to be unique. At write time it will get assigned
+ * an automatic file serial number in the ISO, if needed.
  */
 static
 void catalog_get_id(IsoStream *stream, unsigned int *fs_id, dev_t *dev_id,
@@ -1126,7 +1132,7 @@ int patch_boot_info_table(uint8_t *buf, Ecma119Image *t,
         return iso_msg_submit(t->image->id, ISO_ISOLINUX_CANT_PATCH, 0,
             "Isolinux image too small. We won't patch it.");
     }
-    ret = make_boot_info_table(buf, t->ms_block + (uint32_t) 16,
+    ret = make_boot_info_table(buf, t->opts->ms_block + (uint32_t) 16,
                                t->bootsrc[idx]->sections[0].block,
                                (uint32_t) imgsize);
     return ret;
@@ -1170,10 +1176,10 @@ int iso_patch_eltoritos(Ecma119Image *t)
     continue;
         original = t->bootsrc[idx]->stream;
         size = (size_t) iso_stream_get_size(original);
-
-        /* >>> BOOT ts B00428 :
-               check whether size is not too large for buffering */;
-
+        if (size > Libisofs_elto_max_patchablE)
+            return ISO_PATCH_OVERSIZED_BOOT;
+        if (iso_stream_get_input_stream(original, 0) != NULL)
+            return ISO_PATCH_FILTERED_BOOT;
         buf = calloc(1, size);
         if (buf == NULL) {
             return ISO_OUT_OF_MEM;
@@ -1186,6 +1192,9 @@ int iso_patch_eltoritos(Ecma119Image *t)
         ret = iso_stream_read(original, buf, size);
         iso_stream_close(original);
         if (ret != (int) size) {
+            if (ret >= 0)
+               iso_msg_submit(t->image->id, ISO_FILE_READ_ERROR, 0,
+        "Cannot read all bytes from El Torito boot image for boot info table");
             return (ret < 0) ? ret : (int) ISO_FILE_READ_ERROR;
         }
 
@@ -1302,8 +1311,8 @@ int eltorito_writer_create(Ecma119Image *target)
         }
     }
 
-    if (target->efi_boot_partition != NULL)
-        if (strcmp(target->efi_boot_partition, "--efi-boot-image") == 0)
+    if (target->opts->efi_boot_partition != NULL)
+        if (strcmp(target->opts->efi_boot_partition, "--efi-boot-image") == 0)
             outsource_efi = 1;
     for (idx = 0; idx < target->catalog->num_bootimages; idx++) {
         bootimg = target->catalog->bootimages[idx]->image;
@@ -1337,8 +1346,8 @@ int eltorito_writer_create(Ecma119Image *target)
 
     if (outsource_efi) {
         /* Disable EFI Boot partition and complain */
-        free(target->efi_boot_partition);
-        target->efi_boot_partition = NULL;
+        free(target->opts->efi_boot_partition);
+        target->opts->efi_boot_partition = NULL;
         iso_msg_submit(target->image->id, ISO_BOOT_NO_EFI_ELTO, 0,
 "No newly added El Torito EFI boot image found for exposure as GPT partition");
         return ISO_BOOT_NO_EFI_ELTO;
