@@ -573,6 +573,9 @@ int gzip_clone_stream(IsoStream *old_stream, IsoStream **new_stream, int flag)
 static
 int gzip_cmp_ino(IsoStream *s1, IsoStream *s2);
 
+static
+int gzip_uncompress_cmp_ino(IsoStream *s1, IsoStream *s2);
+
 
 IsoStreamIface gzip_stream_compress_class = {
     4,
@@ -603,17 +606,44 @@ IsoStreamIface gzip_stream_uncompress_class = {
     gzip_stream_free,
     gzip_update_size,
     gzip_get_input_stream,
-    gzip_cmp_ino,
+    gzip_uncompress_cmp_ino,
     gzip_clone_stream
 };
 
-    
+ 
 static
 int gzip_cmp_ino(IsoStream *s1, IsoStream *s2)
 {
+    /* This function may rely on being called by iso_stream_cmp_ino()
+       only with s1, s2 which both point to it as their .cmp_ino() function.
+       It would be a programming error to let any other than 
+       gzip_stream_compress_class point to gzip_cmp_ino().
+       This fallback endangers transitivity of iso_stream_cmp_ino().
+    */
     if (s1->class != s2->class || (s1->class != &gzip_stream_compress_class &&
                                    s2->class != &gzip_stream_compress_class))
         return iso_stream_cmp_ino(s1, s2, 1);
+
+    /* Both streams apply the same treatment to their input streams */
+    return iso_stream_cmp_ino(iso_stream_get_input_stream(s1, 0),
+                              iso_stream_get_input_stream(s2, 0), 0);
+}   
+
+ 
+static
+int gzip_uncompress_cmp_ino(IsoStream *s1, IsoStream *s2)
+{
+    /* This function may rely on being called by iso_stream_cmp_ino()
+       only with s1, s2 which both point to it as their .cmp_ino() function.
+       It would be a programming error to let any other than 
+       gzip_stream_uncompress_class point to gzip_uncompress_cmp_ino().
+    */
+    if (s1->class != s2->class ||
+        (s1->class != &gzip_stream_uncompress_class &&
+         s2->class != &gzip_stream_uncompress_class))
+        return iso_stream_cmp_ino(s1, s2, 1);
+
+    /* Both streams apply the same treatment to their input streams */
     return iso_stream_cmp_ino(iso_stream_get_input_stream(s1, 0),
                               iso_stream_get_input_stream(s2, 0), 0);
 }   
