@@ -115,8 +115,8 @@ uint8_t get_class (uint16_t v)
   return hfsplus_class_pages[high][low];
 }
 
-static
-int set_hfsplus_name(Ecma119Image *t, char *name, HFSPlusNode *node)
+int iso_get_hfsplus_name(char *input_charset, int imgid, char *name,
+                  uint16_t **result, uint32_t *result_len, uint16_t **cmp_name)
 {
     int ret;
     uint16_t *ucs_name, *iptr, *optr;
@@ -128,19 +128,21 @@ int set_hfsplus_name(Ecma119Image *t, char *name, HFSPlusNode *node)
         return ISO_SUCCESS;
     }
 
-    ret = str2utf16be(t->input_charset, name, &ucs_name);
+    ret = str2utf16be(input_charset, name, &ucs_name);
     if (ret < 0) {
-        iso_msg_debug(t->image->id, "Can't convert %s", name);
+        iso_msg_debug(imgid, "Cannot convert '%s'", name);
         return ret;
     }
 
     curlen = ucslen (ucs_name);
-    node->name = calloc ((curlen * HFSPLUS_MAX_DECOMPOSE_LEN + 1),
-			 sizeof (node->name[0]));
-    if (!node->name)
-      return ISO_OUT_OF_MEM;
+    *result = calloc ((curlen * HFSPLUS_MAX_DECOMPOSE_LEN + 1),
+                     sizeof (uint16_t));
+    if (*result == NULL) {
+        free(ucs_name);
+        return ISO_OUT_OF_MEM;
+    }
 
-    for (iptr = ucs_name, optr = node->name; *iptr; iptr++)
+    for (iptr = ucs_name, optr = *result; *iptr; iptr++)
       {
 	const uint16_t *dptr;
 	uint16_t val = iso_ntohs (*iptr);
@@ -189,7 +191,7 @@ int set_hfsplus_name(Ecma119Image *t, char *name, HFSPlusNode *node)
 	if (!ucs_name[0])
 	  break;
 	last_class = get_class (ucs_name[0]);
-	for (optr = node->name + 1; *optr; optr++)
+	for (optr = *result + 1; *optr; optr++)
 	  {
 	    uint8_t new_class = get_class (*optr);
 
@@ -207,11 +209,15 @@ int set_hfsplus_name(Ecma119Image *t, char *name, HFSPlusNode *node)
       }
     while (done);
 
-    node->cmp_name = calloc ((ucslen (node->name) + 1), sizeof (node->cmp_name[0]));
-    if (!node->cmp_name)
-      return ISO_OUT_OF_MEM;
+    *cmp_name = calloc ((ucslen (*result) + 1), sizeof (uint16_t));
+    if (*cmp_name == NULL) {
+        free(ucs_name);
+        free(*result);
+        *result = NULL;
+        return ISO_OUT_OF_MEM;
+    }
 
-    for (iptr = node->name, optr = node->cmp_name; *iptr; iptr++)
+    for (iptr = *result, optr = *cmp_name; *iptr; iptr++)
       {
 	*optr = iso_hfsplus_cichar(*iptr);
 	if (*optr != 0)
@@ -221,10 +227,19 @@ int set_hfsplus_name(Ecma119Image *t, char *name, HFSPlusNode *node)
 
     free (ucs_name);
 
-    node->strlen = ucslen (node->name);
+    *result_len = ucslen (*result);
     return ISO_SUCCESS;
 }
 
+static
+int set_hfsplus_name(Ecma119Image *t, char *name, HFSPlusNode *node)
+{
+    int ret;
+
+    ret = iso_get_hfsplus_name(t->input_charset, t->image->id, name,
+                            &(node->name), &(node->strlen), &(node->cmp_name));
+    return ret;
+}
 
 /* >>> ts B20617
        This should be HFSPlusNode rather than IsoNode in order to have access
@@ -429,7 +444,7 @@ int hfsplus_tail_writer_compute_data_blocks(IsoImageWriter *writer)
   }
 
   t = writer->target;
-  block_size = t->hfsp_block_size;
+  block_size = t->opts->hfsp_block_size;
   block_fac = t->hfsp_iso_block_fac;
 
 #ifdef Libisofs_ts_debuG
@@ -466,7 +481,8 @@ int hfsplus_tail_writer_compute_data_blocks(IsoImageWriter *writer)
 
   t->hfsp_total_blocks = hfsp_curblock - t->hfsp_part_start;
 
-  return iso_quick_apm_entry(t, t->hfsp_part_start / block_fac,
+  return iso_quick_apm_entry(t->apm_req, &(t->apm_req_count),
+                             t->hfsp_part_start / block_fac,
 			     t->hfsp_total_blocks / block_fac +
 			     !!(t->hfsp_total_blocks % block_fac),
                             "HFSPLUS_Hybrid", "Apple_HFS");
@@ -485,7 +501,7 @@ int hfsplus_writer_compute_data_blocks(IsoImageWriter *writer)
     }
 
     t = writer->target;
-    block_size = t->hfsp_block_size;
+    block_size = t->opts->hfsp_block_size;
     block_fac = t->hfsp_iso_block_fac;
 
     iso_msg_debug(t->image->id, "(b) curblock=%d, nodes =%d", t->curblock, t->hfsp_nnodes);
@@ -582,7 +598,7 @@ write_sb (Ecma119Image *t)
 
     iso_msg_debug(t->image->id, "Write HFS+ superblock");
 
-    block_size = t->hfsp_block_size;
+    block_size = t->opts->hfsp_block_size;
 
     memset (buffer, 0, sizeof (buffer));
     ret = iso_write(t, buffer, 1024);
@@ -642,7 +658,7 @@ write_sb (Ecma119Image *t)
 
     }
 
-    memcpy (&sb.num_serial, &t->hfsp_serial_number, 8);
+    memcpy (&sb.num_serial, &t->opts->hfsp_serial_number, 8);
     ret = iso_write(t, &sb, sizeof (sb));
     if (ret < 0)
       return ret;
@@ -665,7 +681,7 @@ int hfsplus_writer_write_data(IsoImageWriter *writer)
     }
 
     t = writer->target;
-    block_size = t->hfsp_block_size;
+    block_size = t->opts->hfsp_block_size;
     block_fac = t->hfsp_iso_block_fac;
     cat_node_size = t->hfsp_cat_node_size;
 
@@ -1033,7 +1049,7 @@ int hfsplus_tail_writer_write_data(IsoImageWriter *writer)
     }
 
     t = writer->target;
-    block_size = t->hfsp_block_size;
+    block_size = t->opts->hfsp_block_size;
 
 #ifdef Libisofs_ts_debuG
     iso_msg_debug(t->image->id, "hfsplus tail writer writes at = %.f",
@@ -1563,7 +1579,7 @@ int mangle_leafs(Ecma119Image *target, int flag)
 int hfsplus_writer_create(Ecma119Image *target)
 {
     int ret;
-    IsoImageWriter *writer;
+    IsoImageWriter *writer = NULL;
     int max_levels;
     int level = 0;
     IsoNode *pos;
@@ -1573,16 +1589,17 @@ int hfsplus_writer_create(Ecma119Image *target)
 
     writer = calloc(1, sizeof(IsoImageWriter));
     if (writer == NULL) {
-        return ISO_OUT_OF_MEM;
+        ret = ISO_OUT_OF_MEM;
+        goto ex;
     }
 
     make_hfsplus_decompose_pages();
     make_hfsplus_class_pages();
 
-    if (target->hfsp_block_size == 0)
-        target->hfsp_block_size = HFSPLUS_DEFAULT_BLOCK_SIZE;
-    target->hfsp_cat_node_size = 2 * target->hfsp_block_size;
-    target->hfsp_iso_block_fac = 2048 / target->hfsp_block_size;
+    if (target->opts->hfsp_block_size == 0)
+        target->opts->hfsp_block_size = HFSPLUS_DEFAULT_BLOCK_SIZE;
+    target->hfsp_cat_node_size = 2 * target->opts->hfsp_block_size;
+    target->hfsp_iso_block_fac = 2048 / target->opts->hfsp_block_size;
     cat_node_size = target->hfsp_cat_node_size;
 
     writer->compute_data_blocks = hfsplus_writer_compute_data_blocks;
@@ -1597,10 +1614,8 @@ int hfsplus_writer_create(Ecma119Image *target)
     target->hfsp_ndirs = 0;
     target->hfsp_cat_id = 16;
     ret = hfsplus_count_tree(target, (IsoNode*)target->image->root);
-    if (ret < 0) {
-        free((char *) writer);
-        return ret;
-    }
+    if (ret < 0)
+        goto ex;
 
     for (i = 0; i < ISO_HFSPLUS_BLESS_MAX; i++)
       target->hfsp_bless_id[i] = 0;
@@ -1610,12 +1625,13 @@ int hfsplus_writer_create(Ecma119Image *target)
 
     target->hfsp_leafs = calloc (target->hfsp_nleafs, sizeof (target->hfsp_leafs[0]));
     if (target->hfsp_leafs == NULL) {
-        return ISO_OUT_OF_MEM;
+        ret = ISO_OUT_OF_MEM;
+        goto ex;
     }
     ret = set_hfsplus_name (target, target->image->volume_id,
                             &target->hfsp_leafs[target->hfsp_curleaf]);
     if (ret < 0)
-        return ret;
+        goto ex;
     target->hfsp_leafs[target->hfsp_curleaf].node = (IsoNode *) target->image->root;
     target->hfsp_leafs[target->hfsp_curleaf].used_size = target->hfsp_leafs[target->hfsp_curleaf].strlen * 2 + 8 + 2 + sizeof (struct hfsplus_catfile_common);
 
@@ -1646,8 +1662,10 @@ int hfsplus_writer_create(Ecma119Image *target)
       {
 	int cret;
 	cret = create_tree(target, pos, 2);
-	if (cret < 0)
-	  return cret;
+	if (cret < 0) {
+	    ret = cret;
+	    goto ex;
+	}
 	pos = pos->next;
 	target->hfsp_leafs[0].nchildren++;
       }
@@ -1657,13 +1675,14 @@ int hfsplus_writer_create(Ecma119Image *target)
 
     ret = mangle_leafs(target, 0);
     if (ret < 0)
-        return ret;
+        goto ex;
 
     for (max_levels = 0; target->hfsp_nleafs >> max_levels; max_levels++);
     max_levels += 2;
     target->hfsp_levels = calloc (max_levels, sizeof (target->hfsp_levels[0]));
     if (target->hfsp_levels == NULL) {
-        return ISO_OUT_OF_MEM;
+	ret = ISO_OUT_OF_MEM;
+	goto ex;
     }
 
     target->hfsp_nnodes = 1;
@@ -1673,9 +1692,10 @@ int hfsplus_writer_create(Ecma119Image *target)
       unsigned bytes_rem = cat_node_size - sizeof (struct hfsplus_btnode) - 2;
 
       target->hfsp_levels[level].nodes = calloc ((target->hfsp_nleafs + 1),  sizeof (target->hfsp_levels[level].nodes[0]));
-      if (!target->hfsp_levels[level].nodes)
-	return ISO_OUT_OF_MEM;
-    
+      if (!target->hfsp_levels[level].nodes) {
+	  ret = ISO_OUT_OF_MEM;
+	  goto ex;
+      }
       target->hfsp_levels[level].level_size = 0;  
       for (i = 0; i < target->hfsp_nleafs; i++)
 	{
@@ -1730,8 +1750,10 @@ int hfsplus_writer_create(Ecma119Image *target)
 	level++;
 
 	target->hfsp_levels[level].nodes = calloc (((last_size + 1) / 2),  sizeof (target->hfsp_levels[level].nodes[0]));
-	if (!target->hfsp_levels[level].nodes)
-	  return ISO_OUT_OF_MEM;
+	if (!target->hfsp_levels[level].nodes) {
+	    ret = ISO_OUT_OF_MEM;
+	    goto ex;
+        }
     
 	target->hfsp_levels[level].level_size = 0;  
 
@@ -1766,16 +1788,21 @@ int hfsplus_writer_create(Ecma119Image *target)
 
     if (target->hfsp_nnodes > (cat_node_size - 0x100) * 8)
       {
-	return iso_msg_submit(target->image->id, ISO_MANGLE_TOO_MUCH_FILES, 0,
-			      "HFS+ map nodes aren't implemented");
-
-	return ISO_MANGLE_TOO_MUCH_FILES;
+	iso_msg_submit(target->image->id, ISO_MANGLE_TOO_MUCH_FILES, 0,
+			"HFS+ map nodes aren't implemented");
+	ret = ISO_MANGLE_TOO_MUCH_FILES;
+	goto ex;
       }
 
     /* add this writer to image */
     target->writers[target->nwriters++] = writer;
+    writer = NULL;
 
-    return ISO_SUCCESS;
+    ret = ISO_SUCCESS;
+ex:;
+    if (writer != NULL)
+        free(writer);
+    return ret;
 }
 
 int hfsplus_tail_writer_create(Ecma119Image *target)
@@ -1822,7 +1849,7 @@ struct iso_hfsplus_xinfo_data *iso_hfsplus_xinfo_new(int flag)
 }
 
 /* The iso_node_xinfo_cloner function which gets associated to
- * iso_hfsplus_xinfo_func by iso_init() resp. iso_init_with_flag() via
+ * iso_hfsplus_xinfo_func by iso_init() or iso_init_with_flag() via
  * iso_node_xinfo_make_clonable()
  */
 int iso_hfsplus_xinfo_cloner(void *old_data, void **new_data, int flag)
