@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
- * Copyright (c) 2009 - 2014 Thomas Schmitt
+ * Copyright (c) 2009 - 2015 Thomas Schmitt
  * 
  * This file is part of the libisofs project; you can redistribute it and/or 
  * modify it under the terms of the GNU General Public License version 2 
@@ -64,6 +64,8 @@ struct Iso_Image
     /* Eventually loaded system area data, or NULL */
     char *system_area_data;
     /* Prescribed/detected options, see iso_write_opts_set_system_area() */
+    /* >>> Needs to be coordinated with .imported_sa_info->system_area_options
+    */
     int system_area_options;
 
    /*
@@ -88,6 +90,9 @@ struct Iso_Image
     char *hppa_kernel_64;
     char *hppa_ramdisk;
 
+    /* Absolute DEC Alpha boot image path in the ISO image */
+    char *alpha_boot_image;
+
     /* image identifier, for message origin identifier */
     int id;
 
@@ -96,6 +101,11 @@ struct Iso_Image
      */
     IsoFilesystem *fs;
 
+    /**
+     * Block storage of imported ISO if demanded by IsoReadOpts.
+     */
+    IsoDataSource *import_src;
+
     /*
      * Default builder to use when adding files to the image tree.
      */
@@ -150,6 +160,20 @@ struct Iso_Image
     /* TODO
     enum iso_replace_mode (*confirm_replace)(IsoFileSource *src, IsoNode *node);
     */
+
+    /**
+     * What to do in case of name longer than truncate_length:
+     *  0= throw FAILURE
+     *  1= truncate to truncate_length with MD5 of whole name at end
+     */
+    int truncate_mode;
+    int truncate_length;
+
+    /**
+     * This is a convenience buffer for name truncation during image
+     * manipulation where libisofs is not thread-safe anyway.
+     */
+    char truncate_buffer[4096];
     
     /**
      * When this is not NULL, it is a pointer to a function that will
@@ -220,6 +244,14 @@ struct Iso_Image
 };
 
 
+/* Apply truncation mode to name, using image->truncate_buffer to perform
+   truncation if needed.
+
+   Warning: Not thread-safe !
+*/
+int iso_image_truncate_name(IsoImage *image, const char *name, char **namept,
+                            int flag);
+
     
 /* Collect the bitmap of used inode numbers in the range of
    _ImageFsData.used_inodes_start + ISO_USED_INODE_RANGE
@@ -385,6 +417,10 @@ struct iso_imported_sys_area {
     char *hppa_ramdisk;
     char *hppa_bootloader;
 
+    uint64_t alpha_boot_image_size;
+    uint64_t alpha_boot_image_adr;
+    char *alpha_boot_image;
+
     /* Some block addresses of active and first session:
          PVD, L Pathtable, Opt L, M Pathtable, Opt M, root directory
     */
