@@ -1,3 +1,13 @@
+/*
+ * Copyright (c) 2008 - 2015 Thomas Schmitt
+ * with special credits to H. Peter Anvin for isohybrid
+ * and to Matthew Garrett for isohybrid with GPT and APM 
+ *
+ * This file is part of the libisofs project; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License version 2
+ * or later as published by the Free Software Foundation.
+ * See COPYING file for details.
+ */
 
 #ifdef HAVE_CONFIG_H
 #include "../config.h"
@@ -50,7 +60,7 @@ license from above stem licenses, typically from LGPL.
 In case its generosity is needed, here is the 2-clause BSD license:
 
 make_isohybrid_mbr.c is copyright 2002-2008 H. Peter Anvin
-                              and 2008-2012 Thomas Schmitt
+                              and 2008-2014 Thomas Schmitt
 
 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
@@ -435,9 +445,10 @@ int assess_isohybrid_gpt_apm(Ecma119Image *t, int *gpt_count, int gpt_idx[128],
                 for (j = 0; j < t->bootsrc[i]->nsections; j++)
                      block_count += t->bootsrc[i]->sections[j].size / 2048;
                 ret = iso_quick_gpt_entry(
-                            t, t->bootsrc[i]->sections[0].block,
-                            block_count, uuid, zero_uuid, gpt_flags,
-                            (uint8_t *) gpt_name);
+                            t->gpt_req, &(t->gpt_req_count),
+                            ((uint64_t) t->bootsrc[i]->sections[0].block) * 4,
+                            ((uint64_t) block_count) * 4,
+                            uuid, zero_uuid, gpt_flags, (uint8_t *) gpt_name);
                 if (ret < 0)
                     return ret;
             }
@@ -449,13 +460,14 @@ int assess_isohybrid_gpt_apm(Ecma119Image *t, int *gpt_count, int gpt_idx[128],
                 block_count = 0;
                 for (j = 0; j < t->bootsrc[i]->nsections; j++)
                      block_count += t->bootsrc[i]->sections[j].size / 2048;
-                ret = iso_quick_apm_entry(t, t->bootsrc[i]->sections[0].block,
+                ret = iso_quick_apm_entry(t->apm_req, &(t->apm_req_count),
+                                          t->bootsrc[i]->sections[0].block,
                                           block_count, "EFI", "Apple_HFS");
                 if (ret < 0)
                     return ret;
                 /* Prevent gap filling */
                 t->apm_req_flags |= 2;
-                t->apm_block_size = 2048;
+                t->opts->apm_block_size = 2048;
             }
         }
     }
@@ -466,7 +478,9 @@ int assess_isohybrid_gpt_apm(Ecma119Image *t, int *gpt_count, int gpt_idx[128],
         iso_ascii_utf_16le(gpt_name);
         /* Let it be open ended. iso_write_gpt() will truncate it as needed. */
         block_count = 0xffffffff;
-        ret = iso_quick_gpt_entry(t, (uint32_t) 0, block_count,
+        ret = iso_quick_gpt_entry(t->gpt_req, &(t->gpt_req_count),
+                                  (uint64_t) t->opts->partition_offset * 4,
+                                  ((uint64_t) block_count) * 4,
                                   basic_data_uuid, zero_uuid, gpt_flags,
                                   (uint8_t *) gpt_name);
         if (ret < 0)
@@ -555,8 +569,9 @@ static int gpt_images_as_mbr_partitions(Ecma119Image *t, char *wpt,
 
 /*
  * @param flag  bit0= make own random MBR Id from current time
+ *              bit1= create protective MBR as of UEFI/GPT specs
  */
-int make_isolinux_mbr(int32_t *img_blocks, Ecma119Image *t,
+int make_isolinux_mbr(uint32_t *img_blocks, Ecma119Image *t,
                       int part_offset, int part_number, int fs_type,
                       uint8_t *buf, int flag)
 {
@@ -570,7 +585,13 @@ int make_isolinux_mbr(int32_t *img_blocks, Ecma119Image *t,
     struct timeval tv;
     struct timezone tz;
 
-    hd_img_blocks = ((off_t) *img_blocks) * (off_t) 4;
+    if (flag & 2) {
+        part_number = 1;
+        part_offset = 1;
+    }
+
+    hd_img_blocks = ((off_t) *img_blocks) * (off_t) 4 -
+                    t->post_iso_part_pad / 512;
 
     boot_lba = t->bootsrc[0]->sections[0].block;
     head_count = t->partition_heads_per_cyl;
@@ -635,14 +656,17 @@ int make_isolinux_mbr(int32_t *img_blocks, Ecma119Image *t,
             wpt+= 16;
     continue;
         }
-        /* write byte 0x80
+        /* write byte 0x80 if bootable
            write LBA_to_CHS(partition_offset)
            write byte filesystem_type
            write LBA_to_CHS(image_size-1)
            write dword partition_offset
            write dword image_size
         */
-        lsb_to_buf(&wpt, 0x80, 8, 0);
+        if (flag & 2)
+            lsb_to_buf(&wpt, 0x00, 8, 0);
+        else
+            lsb_to_buf(&wpt, 0x80, 8, 0);
         lba512chs_to_buf(&wpt, part_offset, head_count, sector_count);
         lsb_to_buf(&wpt, fs_type, 8, 0);
         lba512chs_to_buf(&wpt, hd_img_blocks - 1, head_count, sector_count);
