@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
- * Copyright (c) 2009 - 2011 Thomas Schmitt
+ * Copyright (c) 2009 - 2015 Thomas Schmitt
  *
  * This file is part of the libisofs project; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License version 2 
@@ -38,6 +38,7 @@
 
 #include "util.h"
 #include "node.h"
+#include "stream.h"
 
 
 /*
@@ -232,6 +233,7 @@ void iso_finish()
 {
     libiso_msgs_destroy(&libiso_msgr, 0);
     iso_node_xinfo_dispose_cloners(0);
+    iso_stream_destroy_cmpranks(0);
 }
 
 int iso_set_abort_severity(char *severity)
@@ -503,6 +505,44 @@ const char *iso_error_to_msg(int errcode)
         return "Too many chained symbolic links";
     case ISO_BAD_ISO_FILETYPE:
         return "Unrecognized file type in ISO image";
+    case ISO_NAME_NOT_UCS2:
+        return "Filename not suitable for character set UCS-2";
+    case ISO_IMPORT_COLLISION:
+        return "File name collision during ISO image import";
+    case ISO_HPPA_PALO_INCOMPL:
+        return "Incomplete HP-PA PALO boot parameters";
+    case ISO_HPPA_PALO_OFLOW:
+        return "HP-PA PALO boot address exceeds 2 GB";
+    case ISO_HPPA_PALO_NOTREG:
+        return "HP-PA PALO file is not a data file";
+    case ISO_HPPA_PALO_CMDLEN:
+        return "HP-PA PALO command line too long";
+    case ISO_SYSAREA_PROBLEMS:
+        return "Problems encountered during inspection of System Area";
+    case ISO_INQ_SYSAREA_PROP:
+        return "Unrecognized inquiry for system area property";
+    case ISO_ALPHA_BOOT_NOTREG:
+        return "DEC Alpha Boot Loader file is not a data file";
+    case ISO_NO_KEPT_DATA_SRC:
+        return "No data source of imported ISO image available";
+    case ISO_MALFORMED_READ_INTVL:
+        return "Malformed description string for interval reader";
+    case ISO_INTVL_READ_PROBLEM:
+        return "Unreadable file, premature EOF, or failure to seek for interval reader";
+    case ISO_NOT_REPRODUCIBLE:
+        return "Cannot arrange content of data files in surely reproducible way";
+    case ISO_PATCH_FILTERED_BOOT:
+        return "May not write boot info into filtered stream of boot image";
+    case ISO_PATCH_OVERSIZED_BOOT:
+        return "Boot image to large to buffer for writing boot info";
+    case ISO_RR_NAME_TRUNCATED:
+        return "File name had to be truncated and MD5 marked";
+    case ISO_TRUNCATE_ISOFSNT:
+        return "File name truncation length changed by loaded image info";
+    case ISO_GENERAL_NOTE:
+        return "A general note message was issued";
+    case ISO_BAD_FSRC_FILETYPE:
+        return "Unrecognized file type of IsoFileSrc object";
     default:
         return "Unknown error";
     }
@@ -531,7 +571,8 @@ int iso_msg_submit(int imgid, int errcode, int causedby, const char *fmt, ...)
         vsnprintf(msg, MAX_MSG_LEN, fmt, ap);
         va_end(ap);
     } else {
-        strncpy(msg, iso_error_to_msg(errcode), MAX_MSG_LEN);
+        strncpy(msg, iso_error_to_msg(errcode), MAX_MSG_LEN - 1);
+        msg[MAX_MSG_LEN - 1] = 0;
     }
 
     libiso_msgs_submit(libiso_msgr, imgid, ISO_ERR_CODE(errcode),
