@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
- * Copyright (c) 2009 - 2014 Thomas Schmitt
+ * Copyright (c) 2009 - 2015 Thomas Schmitt
  *
  * This file is part of the libisofs project; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License version 2 
@@ -327,32 +327,50 @@ enum IsoNodeType iso_node_get_type(IsoNode *node)
  * Set the name of a node.
  *
  * @param name  The name in UTF-8 encoding
+ * @param truncate_length (<64 = return on oversized name )
+ * @param flag  bit0= issue warning in case of truncation
  */
-int iso_node_set_name(IsoNode *node, const char *name)
+int iso_node_set_name_trunc(IsoNode *node, const char *in_name,
+                           int truncate_length, int flag)
 {
-    char *new;
+    char *new, *name, *trunc = NULL;
     int ret;
 
     if ((IsoNode*)node->parent == node) {
         /* you can't change name of the root node */
-        return ISO_WRONG_ARG_VALUE;
+        ret = ISO_WRONG_ARG_VALUE;
+        goto ex;
     }
 
+    name = (char *) in_name;
+    if (truncate_length >= 64) {
+        trunc = strdup(name);
+        if (trunc == 0) {
+            ret = ISO_OUT_OF_MEM;
+            goto ex;
+        }
+        ret = iso_truncate_rr_name(1, truncate_length, trunc, !(flag & 1));
+        if (ret < 0)
+            goto ex;
+        name = trunc;
+    }
     /* check if the name is valid */
     ret = iso_node_is_valid_name(name);
     if (ret < 0)
-        return ret;
+        goto ex;
 
     if (node->parent != NULL) {
         /* check if parent already has a node with same name */
         if (iso_dir_get_node(node->parent, name, NULL) == 1) {
-            return ISO_NODE_NAME_NOT_UNIQUE;
+            ret = ISO_NODE_NAME_NOT_UNIQUE;
+            goto ex;
         }
     }
 
     new = strdup(name);
     if (new == NULL) {
-        return ISO_OUT_OF_MEM;
+        ret = ISO_OUT_OF_MEM;
+        goto ex;
     }
     free(node->name);
     node->name = new;
@@ -364,10 +382,29 @@ int iso_node_set_name(IsoNode *node, const char *name)
         iso_node_take(node);
         res = iso_dir_add_node(parent, node, 0);
         if (res < 0) {
-            return res;
+            ret = res;
+            goto ex;
         }
     }
-    return ISO_SUCCESS;
+    ret = ISO_SUCCESS;
+ex:
+    if (trunc != NULL)
+        free(trunc);
+    return ret;
+}
+
+int iso_node_set_name(IsoNode *node, const char *name)
+{
+    return iso_node_set_name_trunc(node, name, 0, 0);
+}
+
+int iso_image_set_node_name(IsoImage *image, IsoNode *node, const char *name,
+                            int flag)
+{
+    if (image->truncate_mode == 0)
+        if ((int) strlen(name) > image->truncate_length)
+            return ISO_RR_NAME_TOO_LONG;
+    return iso_node_set_name_trunc(node, name, image->truncate_length, flag);
 }
 
 /**
@@ -377,6 +414,10 @@ int iso_node_set_name(IsoNode *node, const char *name)
  */
 const char *iso_node_get_name(const IsoNode *node)
 {
+    static char *root = {""};
+
+    if (node->name == NULL)
+        return root;
     return node->name;
 }
 
@@ -610,6 +651,43 @@ int iso_dir_get_node(IsoDir *dir, const char *name, IsoNode **node)
     return 1;
 }
 
+int iso_dir_get_node_trunc(IsoDir *dir, int truncate_length,
+                           const char *name, IsoNode **node)
+{
+    int ret;
+    char *trunc = NULL;
+
+    if ((int) strlen(name) <= truncate_length) {
+        ret = iso_dir_get_node(dir, name, node);
+        return ret;
+    }
+    trunc = strdup(name);
+    if (trunc == NULL)
+        return ISO_OUT_OF_MEM;
+    ret = iso_truncate_rr_name(1, truncate_length, trunc, 1);
+    if (ret < 0)
+        goto ex;
+    ret = iso_dir_get_node(dir, trunc, node);
+    if (ret == 0)
+        ret = 2;
+ex:;
+    LIBISO_FREE_MEM(trunc);
+    return ret;
+}
+
+/* API */
+int iso_image_dir_get_node(IsoImage *image, IsoDir *dir,
+                           const char *name, IsoNode **node, int flag)
+{
+    int ret;
+
+    if (image->truncate_mode == 0 || (flag & 1))
+        ret = iso_dir_get_node(dir, name, node);
+    else
+        ret = iso_dir_get_node_trunc(dir, image->truncate_length, name, node);
+    return ret;
+}
+
 /**
  * Get the number of children of a directory.
  *
@@ -1117,16 +1195,17 @@ int iso_file_get_old_image_lba(IsoFile *file, uint32_t *lba, int flag)
 {
     int ret;
     int section_count;
-    struct iso_file_section *sections;
+    struct iso_file_section *sections = NULL;
+
     if (file == NULL || lba == NULL) {
         return ISO_NULL_POINTER;
     }
-    ret = iso_file_get_old_image_sections(file, &section_count, &sections, flag);
-    if (ret <= 0) {
+    ret = iso_file_get_old_image_sections(file, &section_count, &sections, 0);
+    if (ret <= 0)
         return ret;
-    }
     if (section_count != 1) {
-        free(sections);
+        if (sections != NULL)
+            free(sections);
         return ISO_WRONG_ARG_VALUE;
     }
     *lba = sections[0].block;
@@ -1135,7 +1214,6 @@ int iso_file_get_old_image_lba(IsoFile *file, uint32_t *lba, int flag)
 }
 
 
-
 /*
  * Like iso_file_get_old_image_lba(), but take an IsoNode.
  *
@@ -1531,11 +1609,17 @@ int attrs_cleanout_name(char *del_name, size_t *num_attrs, char **names,
     size_t i, w;
 
     for (w = i = 0; i < *num_attrs; i++) {
-        if ((strcmp(names[i], del_name) == 0) ^ (flag & 1))
-            continue;
+        if ((strcmp(names[i], del_name) == 0) ^ (flag & 1)) {
+            if (names[i] != NULL)
+                free(names[i]);
+            if (values[i] != NULL)
+                free(values[i]);
+             names[i] = values[i] = NULL;
+    continue;
+        }
         if (w == i) {
             w++;
-            continue;
+    continue;
         }
         names[w] = names[i];
         value_lengths[w] = value_lengths[i];
@@ -1881,8 +1965,9 @@ int iso_node_set_attrs(IsoNode *node, size_t num_attrs, char **names,
                        size_t *value_lengths, char **values, int flag)
 {
     int ret, acl_saved = 0;
-    size_t sret, result_len, m_num = 0, *m_value_lengths = NULL, i;
-    unsigned char *result;
+    ssize_t sret;
+    size_t result_len, m_num = 0, *m_value_lengths = NULL, i;
+    unsigned char *result = NULL;
     char *a_acl = NULL, *d_acl = NULL, **m_names = NULL, **m_values = NULL;
 
     if (!(flag & 8))
@@ -1921,28 +2006,34 @@ int iso_node_set_attrs(IsoNode *node, size_t num_attrs, char **names,
     }
     sret = aaip_encode(num_attrs, names, value_lengths, values,
                        &result_len, &result, 0);
-    if (sret == 0) {
-        ret = ISO_OUT_OF_MEM;
+    if (sret < 0) {
+        ret = sret;
         goto ex;
     }
 
     ret = iso_node_remove_xinfo(node, aaip_xinfo_func);
-    if (ret < 0)
-        goto ex;
-    ret = iso_node_add_xinfo(node, aaip_xinfo_func, result);
-    if (ret < 0)
-        goto ex;
-    if (ret == 0) {
-
-        /* >>> something is messed up with xinfo: an aa_string still exists */;
-
-        ret = ISO_ERROR;
+    if (ret < 0) {
+        if (result != NULL)
+            free(result);
         goto ex;
     }
-    if (acl_saved) {
-        ret = iso_node_set_acl_text(node, a_acl, d_acl, 0);
+    if (sret > 0) {
+        ret = iso_node_add_xinfo(node, aaip_xinfo_func, result);
         if (ret < 0)
             goto ex;
+        if (ret == 0) {
+
+            /* >>> something is messed up with xinfo:
+                   an aa_string still exists */;
+
+            ret = ISO_ERROR;
+            goto ex;
+        }
+        if (acl_saved) {
+            ret = iso_node_set_acl_text(node, a_acl, d_acl, 0);
+            if (ret < 0)
+                goto ex;
+        }
     }
     ret = 1;
 ex:;
@@ -2160,10 +2251,8 @@ int iso_node_set_acl_text(IsoNode *node, char *access_text, char *default_text,
                 ret = ISO_AAIP_BAD_ACL_TEXT;
                 goto ex;
             }
-            ret = 1;
-            if (a_text != NULL || d_text != NULL)
-                ret = aaip_encode_both_acl(a_text, d_text, st_mode,
-                                           &acl_len, &acl, 2 | 8);
+            ret = aaip_encode_both_acl(a_text, d_text, st_mode,
+                                       &acl_len, &acl, 2 | 8);
         } else {
             ret = 1;
             if (access_text != NULL || default_text != NULL)
@@ -2868,6 +2957,50 @@ ex:;
 }
 
 
+int iso_root_set_isofsnt(IsoNode *node, uint32_t truncate_mode,
+                         uint32_t truncate_length, int flag)
+{
+    char buffer[5 + 5], *wpt = buffer, *valuept = buffer;
+    int result_len, ret;
+    static char *names = "isofs.nt";
+    static size_t value_lengths[1];
+
+    iso_util_encode_len_bytes(truncate_mode, wpt, 0, &result_len, 0);
+    wpt += result_len;
+    iso_util_encode_len_bytes(truncate_length, wpt, 0, &result_len, 0);
+    wpt += result_len;
+    value_lengths[0] = wpt - buffer;
+    ret = iso_node_set_attrs(node, (size_t) 1,
+                             &names, value_lengths, &valuept, 2 | 8);
+    return ret;
+}
+
+
+int iso_root_get_isofsnt(IsoNode *node, uint32_t *truncate_mode,
+                         uint32_t *truncate_length, int flag)
+{
+    int ret, len;
+    size_t value_len;
+    char *value = NULL, *rpt;
+
+    ret = iso_node_lookup_attr(node, "isofs.nt", &value_len, &value, 0);
+    if (ret <= 0)
+        goto ex;
+
+    rpt = value;
+    iso_util_decode_len_bytes(truncate_mode, rpt, &len,
+                              value_len - (rpt - value), 0);
+    rpt += len + 1;
+    iso_util_decode_len_bytes(truncate_length, rpt, &len,
+                              value_len - (rpt - value), 0);
+    ret= ISO_SUCCESS;
+ex:;
+    if (value != NULL)
+        free(value);
+    return ret;
+}
+
+
 /* API */
 int iso_file_get_md5(IsoImage *image, IsoFile *file, char md5[16], int flag)
 {
@@ -2921,21 +3054,23 @@ int iso_file_make_md5(IsoFile *file, int flag)
 
     if (file->from_old_session)
         dig = 1;
-    md5= calloc(16, 1);
+    md5 = calloc(16, 1);
+    if (md5 == NULL) 
+        return ISO_OUT_OF_MEM;
     ret = iso_stream_make_md5(file->stream, md5, dig);
-    if (ret < 0)
-        goto ex;
+    if (ret < 0) {
+        free(md5);
+        return ret;
+    }
     iso_node_remove_xinfo((IsoNode *) file, checksum_md5_xinfo_func);
     ret = iso_node_add_xinfo((IsoNode *) file, checksum_md5_xinfo_func, md5);
     if (ret == 0)
         ret = ISO_ERROR; /* should not happen after iso_node_remove_xinfo() */
     if (ret < 0) {
         free(md5);
-        goto ex;
+        return ret;
     }
-    ret = 1;
-ex:;
-    return ret;
+    return 1;
 }
 
 
