@@ -512,6 +512,16 @@ int iso_root_get_isofsca(IsoNode *node, uint32_t *start_lba, uint32_t *end_lba,
                          int flag);
 
 
+/**
+ * Record and get truncation parameters as of iso_image_set_truncate_mode() by
+ * "isofs.nt".
+ */
+int iso_root_set_isofsnt(IsoNode *node, uint32_t truncate_mode,
+                         uint32_t truncate_length, int flag);
+int iso_root_get_isofsnt(IsoNode *node, uint32_t *truncate_mode,
+                         uint32_t *truncate_length, int flag);
+
+
 /**
  * Copy the xinfo list from one node to the another.
  */
@@ -543,4 +553,10 @@ int zisofs_zf_xinfo_func(void *data, int flag);
 int zisofs_zf_xinfo_cloner(void *old_data, void **new_data, int flag);
 
 
+/* Performing search for possibly truncated node name.
+ */
+int iso_dir_get_node_trunc(IsoDir *dir, int truncate_length,
+                           const char *name, IsoNode **node);
+
+
 #endif /*LIBISO_NODE_H_*/
