@@ -3,8 +3,9 @@
  * Copyright (c) 2009 Thomas Schmitt
  * 
  * This file is part of the libisofs project; you can redistribute it and/or 
- * modify it under the terms of the GNU General Public License version 2 as 
- * published by the Free Software Foundation. See COPYING file for details.
+ * modify it under the terms of the GNU General Public License version 2 
+ * or later as published by the Free Software Foundation. 
+ * See COPYING file for details.
  */
 
 /*
@@ -12,6 +13,10 @@
  * Rock Ridge and AAIP extensions on an ECMA-119 image.
  */
 
+#ifdef HAVE_CONFIG_H
+#include "../config.h"
+#endif
+
 #include "libisofs.h"
 #include "ecma119.h"
 #include "util.h"
@@ -76,8 +81,7 @@ int susp_iter_next(SuspIterator *iter, struct susp_sys_user_entry **sue)
          * (IEEE 1281, SUSP. section 4) 
          */
         if (iter->ce_len) {
-            uint32_t block;
-            int nblocks;
+            uint32_t block, nblocks;
 
             /* A CE has found, there is another continuation area */
             nblocks = DIV_UP(iter->ce_off + iter->ce_len, BLOCK_SIZE);
@@ -207,10 +211,19 @@ int read_rr_TF(struct susp_sys_user_entry *tf, struct stat *st)
     
     /* 1. Creation time */
     if (tf->data.TF.flags[0] & (1 << 0)) {
-        
-        /* the creation is the recording time. we ignore this */
-        /* TODO maybe it would be good to manage it in ms discs, where
-         *      the recording time could be different than now!! */     
+        /* Linux accepts ctime by Creation time and by Attributes time.
+         * If both are given, then Attribute time will win.
+         */
+        if (tf->len_sue[0] < 5 + (nts+1) * s) {
+            /* RR TF entry too short. */
+            return ISO_WRONG_RR;
+        }
+        if (s == 7) {
+            time = iso_datetime_read_7(&tf->data.TF.t_stamps[nts*7]);
+        } else {
+            time = iso_datetime_read_17(&tf->data.TF.t_stamps[nts*17]);
+        }
+        st->st_ctime = time;
         ++nts;
     }
     
@@ -408,7 +421,7 @@ int read_rr_PN(struct susp_sys_user_entry *pn, struct stat *st)
 {
     int high_shift= 0;
 
-    if (pn == NULL || pn == NULL) {
+    if (pn == NULL || st == NULL) {
         return ISO_NULL_POINTER;
     }
     if (pn->sig[0] != 'P' || pn->sig[1] != 'N') {
@@ -438,7 +451,7 @@ int read_rr_PN(struct susp_sys_user_entry *pn, struct stat *st)
 }
 
 
-/* AA is the field signature of AAIP versions < 2.0
+/* AA is the obsolete field signature of AAIP versions < 2.0
 */
 int read_aaip_AA(struct susp_sys_user_entry *sue,
                  unsigned char **aa_string, size_t *aa_size, size_t *aa_len,
@@ -509,7 +522,7 @@ int read_aaip_AA(struct susp_sys_user_entry *sue,
 }
 
 
-/* AL is the obsolete field signature of AAIP versions >= 2.0
+/* AL is the field signature of AAIP versions >= 2.0
 */
 int read_aaip_AL(struct susp_sys_user_entry *sue,
                  unsigned char **aa_string, size_t *aa_size, size_t *aa_len,
