@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
- * Copyright (c) 2011 Thomas Schmitt
+ * Copyright (c) 2011 - 2015 Thomas Schmitt
  * 
  * This file is part of the libisofs project; you can redistribute it and/or 
  * modify it under the terms of the GNU General Public License version 2 
@@ -23,6 +23,7 @@
 #include "builder.h"
 #include "messages.h"
 #include "tree.h"
+#include "util.h"
 
 #include <stdlib.h>
 #include <string.h>
@@ -99,6 +100,19 @@ int iso_tree_add_new_dir(IsoDir *parent, const char *name, IsoDir **dir)
     return iso_dir_insert(parent, (IsoNode*)node, pos, ISO_REPLACE_NEVER);
 }
 
+int iso_image_add_new_dir(IsoImage *image, IsoDir *parent, const char *name,
+                          IsoDir **dir)
+{
+    int ret;
+    char *namept;
+
+    ret = iso_image_truncate_name(image, name, &namept, 0);
+    if (ret < 0)
+        return ret;
+    ret = iso_tree_add_new_dir(parent, namept, dir);
+    return ret;
+}
+
 /**
  * Add a new symlink to the directory tree. Permissions are set to 0777, 
  * owner and hidden atts are taken from parent. You can modify any of them 
@@ -174,6 +188,20 @@ int iso_tree_add_new_symlink(IsoDir *parent, const char *name,
     return iso_dir_insert(parent, (IsoNode*)node, pos, ISO_REPLACE_NEVER);
 }
 
+int iso_image_add_new_symlink(IsoImage *image, IsoDir *parent,
+                              const char *name, const char *dest,
+                              IsoSymlink **link)
+{
+    int ret;
+    char *namept;
+
+    ret = iso_image_truncate_name(image, name, &namept, 0);
+    if (ret < 0)
+        return ret;
+    ret = iso_tree_add_new_symlink(parent, namept, dest, link);
+    return ret;
+}
+
 /**
  * Add a new special file to the directory tree. As far as libisofs concerns,
  * an special file is a block device, a character device, a FIFO (named pipe)
@@ -263,6 +291,20 @@ int iso_tree_add_new_special(IsoDir *parent, const char *name, mode_t mode,
     return iso_dir_insert(parent, (IsoNode*)node, pos, ISO_REPLACE_NEVER);
 }
 
+int iso_image_add_new_special(IsoImage *image, IsoDir *parent,
+                              const char *name, mode_t mode,
+                              dev_t dev, IsoSpecial **special)
+{
+    int ret;
+    char *namept;
+
+    ret = iso_image_truncate_name(image, name, &namept, 0);
+    if (ret < 0)
+        return ret;
+    ret = iso_tree_add_new_special(parent, namept, mode, dev, special);
+    return ret;
+}
+
 /**
  * Add a new regular file to the iso tree. Permissions are set to 0444, 
  * owner and hidden atts are taken from parent. You can modify any of them 
@@ -338,6 +380,19 @@ int iso_tree_add_new_file(IsoDir *parent, const char *name, IsoStream *stream,
     return iso_dir_insert(parent, (IsoNode*)node, pos, ISO_REPLACE_NEVER);
 }
 
+int iso_image_add_new_file(IsoImage *image, IsoDir *parent, const char *name,
+                           IsoStream *stream, IsoFile **file)
+{
+    int ret;
+    char *namept;
+
+    ret = iso_image_truncate_name(image, name, &namept, 0);
+    if (ret < 0)
+        return ret;
+    ret = iso_tree_add_new_file(parent, namept, stream, file);
+    return ret;
+}
+
 /**
  * Set whether to follow or not symbolic links when added a file from a source
  * to IsoImage.
@@ -502,10 +557,10 @@ int iso_tree_add_node_builder(IsoImage *image, IsoDir *parent,
     int result;
     IsoNode *new;
     IsoNode **pos;
-    char *name;
+    char *name = NULL, *namept;
 
     if (parent == NULL || src == NULL || builder == NULL) {
-        return ISO_NULL_POINTER;
+        result = ISO_NULL_POINTER; goto ex;
     }
     if (node) {
         *node = NULL;
@@ -513,25 +568,31 @@ int iso_tree_add_node_builder(IsoImage *image, IsoDir *parent,
 
     name = iso_file_source_get_name(src);
 
+    result = iso_image_truncate_name(image, name, &namept, 0);
+    if (result < 0)
+        return result;
+
     /* find place where to insert */
-    result = iso_dir_exists(parent, name, &pos);
-    free(name);
+    result = iso_dir_exists(parent, namept, &pos);
     if (result) {
         /* a node with same name already exists */
-        return ISO_NODE_NAME_NOT_UNIQUE;
+        result = ISO_NODE_NAME_NOT_UNIQUE; goto ex;
     }
 
-    result = builder->create_node(builder, image, src, &new);
-    if (result < 0) {
-        return result;
-    }
+    result = builder->create_node(builder, image, src, namept, &new);
+    if (result < 0)
+        goto ex;
 
     if (node) {
         *node = new;
     }
 
     /* finally, add node to parent */
-    return iso_dir_insert(parent, (IsoNode*)new, pos, ISO_REPLACE_NEVER);
+    result = iso_dir_insert(parent, (IsoNode*)new, pos, ISO_REPLACE_NEVER);
+ex:
+    if (name != NULL)
+        free(name);
+    return result;
 }
 
 int iso_tree_add_node(IsoImage *image, IsoDir *parent, const char *path,
@@ -565,6 +626,7 @@ int iso_tree_add_new_node(IsoImage *image, IsoDir *parent, const char *name,
     IsoFileSource *file;
     IsoNode *new;
     IsoNode **pos;
+    char *namept;
 
     if (image == NULL || parent == NULL || name == NULL || path == NULL) {
         return ISO_NULL_POINTER;
@@ -574,8 +636,12 @@ int iso_tree_add_new_node(IsoImage *image, IsoDir *parent, const char *name,
         *node = NULL;
     }
 
+    result = iso_image_truncate_name(image, name, &namept, 0);
+    if (result < 0)
+        return result;
+
     /* find place where to insert */
-    result = iso_dir_exists(parent, name, &pos);
+    result = iso_dir_exists(parent, namept, &pos);
     if (result) {
         /* a node with same name already exists */
         return ISO_NODE_NAME_NOT_UNIQUE;
@@ -587,7 +653,8 @@ int iso_tree_add_new_node(IsoImage *image, IsoDir *parent, const char *name,
         return result;
     }
 
-    result = image->builder->create_node(image->builder, image, file, &new);
+    result = image->builder->create_node(image->builder, image, file,
+                                         namept, &new);
     
     /* free the file */
     iso_file_source_unref(file);
@@ -596,12 +663,6 @@ int iso_tree_add_new_node(IsoImage *image, IsoDir *parent, const char *name,
         return result;
     }
     
-    result = iso_node_set_name(new, name);
-    if (result < 0) {
-        iso_node_unref(new);
-        return result;
-    }
-
     if (node) {
         *node = new;
     }
@@ -622,6 +683,7 @@ int iso_tree_add_new_cut_out_node(IsoImage *image, IsoDir *parent,
     IsoFile *new;
     IsoNode **pos;
     IsoStream *stream;
+    char *namept;
 
     if (image == NULL || parent == NULL || name == NULL || path == NULL) {
         return ISO_NULL_POINTER;
@@ -631,8 +693,12 @@ int iso_tree_add_new_cut_out_node(IsoImage *image, IsoDir *parent,
         *node = NULL;
     }
 
+    result = iso_image_truncate_name(image, name, &namept, 0);
+    if (result < 0)
+        return result;
+
     /* find place where to insert */
-    result = iso_dir_exists(parent, name, &pos);
+    result = iso_dir_exists(parent, namept, &pos);
     if (result) {
         /* a node with same name already exists */
         return ISO_NODE_NAME_NOT_UNIQUE;
@@ -675,7 +741,7 @@ int iso_tree_add_new_cut_out_node(IsoImage *image, IsoDir *parent,
     iso_stream_unref(new->stream);
     new->stream = stream;
     
-    result = iso_node_set_name((IsoNode*)new, name);
+    result = iso_node_set_name((IsoNode*)new, namept);
     if (result < 0) {
         iso_node_unref((IsoNode*)new);
         return result;
@@ -742,6 +808,176 @@ int check_special(IsoImage *image, mode_t mode)
     return 0;
 }
 
+
+static
+void ascii_increment(char *name, int len, int pos, int rollover_carry)
+{
+     int c;
+
+again:;
+     if (pos < 0 || pos >= len)
+         pos = len - 1;
+     c = name[pos];
+     if (c >= '0' && c < '9') {
+         c++;
+     } else if (c == '9') {
+         c = 'A';
+     } else if (c >= 'A' && c < 'Z') {
+         c++;
+     } else if (c == 'Z') {
+         c = '_';
+     } else if (c == '_') {
+         c = 'a';
+     } else if (c >= 'a' && c < 'z') {
+         c++;
+     } else if (c == 'z') {
+         c = '0';
+         name[pos] = c;
+         pos--;
+         if (pos >= 0 || rollover_carry)
+             goto again;
+         return;
+     } else {
+         if (pos == len - 1 || name[pos + 1] == '.')
+             c = '_'; /* Make first change less riddling */
+         else
+             c = '0'; /* But else use the full range of valid characters */
+     }
+     name[pos] = c;
+}
+
+static
+int insert_underscores(char *name, int *len, int *at_pos, int count,
+                       char **new_name)
+{
+    int ret;
+
+    LIBISO_ALLOC_MEM(*new_name, char, count + *len + 1);
+    if (*at_pos > 0)
+        memcpy(*new_name, name, *at_pos);
+    if (count > 0)
+        memset(*new_name + *at_pos, '_', count);
+    if (*len > *at_pos)
+        memcpy(*new_name + *at_pos + count, name + *at_pos, *len - *at_pos);
+    (*new_name)[count + *len] = 0;
+    *len += count;
+    *at_pos += count;
+    ret= ISO_SUCCESS;
+ex:;
+    return ret;
+}
+
+static
+int make_incrementable_name(char **name, char **unique_name, int *low_pos,
+                            int *rollover_carry, int *pre_check) 
+{
+    char *dpt, *npt;
+    int first, len, ret;
+    
+    /* The incrementable part of the file shall have at least 7 characters.
+       There may be up to pow(2.0,32.0)*2048/33 = 266548273400 files.
+       The set of increment result characters has 63 elements.
+       pow(63.0,7.0) is nearly 15 times larger than 266548273400.
+    */
+    static int min_incr = 7;
+
+    /* At most two suffixes of total length up to 12, like .tar.bz2,
+       shall be preserved. The incrementable part will eventually be
+       padded up.
+       Incrementing begins before the last suffix in any case. But when this
+       rolls over on short prefixes, then long last suffixes will get used
+       as high characters of the incremental part. This is indicated by
+       *rollover_carry which corresponds to the parameter of ascii_increment()
+       with the same name.
+    */
+    static int max_suffix = 12;
+
+    *rollover_carry = 0;
+    *pre_check = 0;
+
+    len = strlen(*name);
+
+    /* Check if the part before the first dot is long enough.
+       If not, then preserve the last two short suffixes.
+    */
+    dpt = strchr(*name, '.');
+    if (dpt != NULL)
+        if ((dpt - *name) < min_incr)
+            dpt = strrchr(*name, '.');
+    if (dpt != NULL) {
+        first= (dpt - *name);
+        if (dpt > *name && len - first < max_suffix) {
+            for(npt = dpt - 1; npt >= *name && *npt != '.'; npt--);
+            if (npt >= *name) {
+                if (len - (npt - *name) <= max_suffix) {
+                    first= (npt - *name);
+                    dpt = npt;
+                }
+            }
+        }
+    } else
+        first= len;
+    if (first < min_incr && (len - first) <= max_suffix) {
+        ret = insert_underscores(*name, &len, &first, min_incr - first,
+                                 unique_name);
+        if (ret < 0)
+            goto ex;
+        *pre_check = 1; /* It might now already be unique */
+
+    } else if (len < 64) {
+        /* Insert an underscore to preserve the original name at least for the
+           first few increments
+        */
+        ret = insert_underscores(*name, &len, &first, 1, unique_name);
+        if (ret < 0)
+            goto ex;
+        *pre_check = 1;
+
+    } else {
+        LIBISO_ALLOC_MEM(*unique_name, char, len + 1);
+        memcpy(*unique_name, *name, len);
+        if (first < min_incr)
+            *rollover_carry = 1; /* Do not get caged before the dots */
+    }
+    (*unique_name)[len] = 0;
+    *low_pos = first - 1;
+    ret = 1;
+ex:;
+    return(ret);
+}
+
+static
+int make_really_unique_name(IsoDir *parent, char **name, char **unique_name,
+                            IsoNode ***pos, int flag)
+{
+    int ret, rollover_carry = 0, pre_check = 0, ascii_idx = -1, len;
+
+    ret = make_incrementable_name(name, unique_name, &ascii_idx,
+                                  &rollover_carry, &pre_check);
+    if (ret < 0)
+        goto ex;
+    len = strlen(*unique_name);
+    while (1) {
+        if (!pre_check)
+            ascii_increment(*unique_name, len, ascii_idx, !!rollover_carry);
+        else
+            pre_check = 0;
+        ret = iso_dir_exists(parent, *unique_name, pos);
+        if (ret < 0)
+            goto ex;
+        if (ret == 0)
+    break;
+    }
+    *name = *unique_name;
+    ret = ISO_SUCCESS;
+ex:;
+    if (ret < 0) {
+        LIBISO_FREE_MEM(*unique_name);
+        *unique_name = NULL;
+    }
+    return ret;
+}
+
 /**
  * Recursively add a given directory to the image tree.
  * 
@@ -750,12 +986,12 @@ int check_special(IsoImage *image, mode_t mode)
  */
 int iso_add_dir_src_rec(IsoImage *image, IsoDir *parent, IsoFileSource *dir)
 {
-    int ret;
+    int ret, dir_is_open = 0;
     IsoNodeBuilder *builder;
     IsoFileSource *file;
     IsoNode **pos;
     struct stat info;
-    char *name, *path;
+    char *name, *path, *allocated_name = NULL;
     IsoNode *new;
     enum iso_replace_mode replace;
 
@@ -771,8 +1007,9 @@ int iso_add_dir_src_rec(IsoImage *image, IsoDir *parent, IsoFileSource *dir)
             ret = iso_msg_submit(image->id, ISO_NULL_POINTER, ret,
                            "Can't open dir. NULL pointer caught as dir name");
         }
-        return ret;
+        goto ex;
     }
+    dir_is_open = 1;
 
     builder = image->builder;
     
@@ -785,15 +1022,16 @@ int iso_add_dir_src_rec(IsoImage *image, IsoDir *parent, IsoFileSource *dir)
             if (ret < 0) {
                 /* error reading dir */
                 ret = iso_msg_submit(image->id, ret, ret, "Error reading dir");
+                goto ex;
             }
-            break;
+    break; /* End of directory */
         }
 
         path = iso_file_source_get_path(file);
         if (path == NULL) {
             ret = iso_msg_submit(image->id, ISO_NULL_POINTER, ret, 
                                  "NULL pointer caught as file path");
-            return ret;
+            goto ex;
         }
         name = strrchr(path, '/') + 1;
 
@@ -827,19 +1065,25 @@ int iso_add_dir_src_rec(IsoImage *image, IsoDir *parent, IsoFileSource *dir)
 
         /* find place where to insert */
         ret = iso_dir_exists(parent, name, &pos);
-        /* TODO
-         * if (ret && replace == ISO_REPLACE_ASK) {
-         *    replace = /....
-         * }
-         */
-        
-        /* chek if we must insert or not */
-        /* TODO check for other replace behavior */
-        if (ret && (replace == ISO_REPLACE_NEVER)) { 
-            /* skip file */
-            goto dir_rec_continue;
+        if (ret) {
+            /* Resolve name collision
+               e.g. caused by fs_image.c:make_hopefully_unique_name() 
+            */
+            LIBISO_FREE_MEM(allocated_name); allocated_name = NULL;
+            ret = make_really_unique_name(parent, &name, &allocated_name, &pos,
+                                          0);
+            if (ret < 0)
+                goto ex;
+            image->collision_warnings++;
+            if (image->collision_warnings < ISO_IMPORT_COLL_WARN_MAX) {
+                ret = iso_msg_submit(image->id, ISO_IMPORT_COLLISION, 0, 
+                         "File name collision resolved with %s . Now: %s",
+                         path, name);
+                if (ret < 0)
+                    goto ex;
+            }
         }
-        
+
         /* if we are here we must insert. Give user a chance for cancel */
         if (image->report) {
             int r = image->report(image, file);
@@ -848,7 +1092,7 @@ int iso_add_dir_src_rec(IsoImage *image, IsoDir *parent, IsoFileSource *dir)
                 goto dir_rec_continue;
             }
         }
-        ret = builder->create_node(builder, image, file, &new);
+        ret = builder->create_node(builder, image, file, name, &new);
         if (ret < 0) {
             ret = iso_msg_submit(image->id, ISO_FILE_CANT_ADD, ret,
                          "Error when adding file %s", path);
@@ -884,14 +1128,17 @@ dir_rec_continue:;
         /* check for error severity to decide what to do */
         if (ret < 0) {
             ret = iso_msg_submit(image->id, ret, 0, NULL);
-            if (ret < 0) {
-                break;
-            }
+            if (ret < 0)
+                goto ex;
         }
     } /* while */
 
-    iso_file_source_close(dir);
-    return ret < 0 ? ret : ISO_SUCCESS;
+    ret = ISO_SUCCESS;
+ex:;
+    if (dir_is_open)
+        iso_file_source_close(dir);
+    LIBISO_FREE_MEM(allocated_name);
+    return ret;
 }
 
 int iso_tree_add_dir_rec(IsoImage *image, IsoDir *parent, const char *dir)
@@ -927,7 +1174,10 @@ int iso_tree_add_dir_rec(IsoImage *image, IsoDir *parent, const char *dir)
     return result;
 }
 
-int iso_tree_path_to_node(IsoImage *image, const char *path, IsoNode **node)
+/* @param flag bit0= truncate according to image truncate mode and length
+*/
+int iso_tree_path_to_node_flag(IsoImage *image, const char *path,
+                               IsoNode **node, int flag)
 {
     int result;
     IsoNode *n;
@@ -949,6 +1199,8 @@ int iso_tree_path_to_node(IsoImage *image, const char *path, IsoNode **node)
     }
 
     ptr = strdup(path);
+    if (ptr == NULL)
+        return ISO_OUT_OF_MEM;
     result = 0;
 
     /* get the first component of the path */
@@ -961,7 +1213,12 @@ int iso_tree_path_to_node(IsoImage *image, const char *path, IsoNode **node)
         }
         dir = (IsoDir *)n;
 
-        result = iso_dir_get_node(dir, component, &n);
+        if ((flag & 1) && image->truncate_mode == 1) {
+            result = iso_dir_get_node_trunc(dir, image->truncate_length,
+                                            component, &n);
+        } else {
+            result = iso_dir_get_node(dir, component, &n);
+        }
         if (result != 1) {
             n = NULL;
             break;
@@ -977,6 +1234,16 @@ int iso_tree_path_to_node(IsoImage *image, const char *path, IsoNode **node)
     return result;
 }
 
+int iso_tree_path_to_node(IsoImage *image, const char *path, IsoNode **node)
+{
+    return iso_tree_path_to_node_flag(image, path, node, 0);
+}
+
+int iso_image_path_to_node(IsoImage *image, const char *path, IsoNode **node)
+{
+    return iso_tree_path_to_node_flag(image, path, node, 1);
+}
+
 char *iso_tree_get_node_path(IsoNode *node)
 {
     char *path = NULL, *parent_path = NULL;
@@ -1008,6 +1275,65 @@ ex:;
     return path;
 }
 
+/* Note: No reference is taken to the found node.
+   @param flag bit0= recursion
+*/
+int iso_tree_get_node_of_block(IsoImage *image, IsoDir *dir, uint32_t block,
+                               IsoNode **found, uint32_t *next_above, int flag)
+{
+    int ret, section_count, i;
+    IsoDirIter *iter = NULL;
+    IsoNode *node;
+    IsoDir *subdir;
+    IsoFile *file;
+    struct iso_file_section *sections = NULL;
+    uint32_t na = 0;
+
+    if (dir == NULL)
+        dir = image->root;
+
+    ret = iso_dir_get_children(dir, &iter);
+    while (iso_dir_iter_next(iter, &node) == 1 ) {
+
+        if (ISO_NODE_IS_FILE(node)) {
+            file = (IsoFile *) node;
+            ret = iso_file_get_old_image_sections(file, &section_count,
+                                                  &sections, 0);
+            if (ret <= 0)
+    continue;
+            for (i = 0; i < section_count; i++) {
+                if (sections[i].block <= block &&
+                    block - sections[i].block <
+                                  (((off_t) sections[i].size) + 2047) / 2048) {
+                    *found = node;
+                    ret = 1; goto ex;
+                }
+                if ((na == 0 || sections[i].block < na) &&
+                                                     sections[i].block > block)
+                    na = sections[i].block;
+            }
+            free(sections); sections = NULL;
+        } else if (ISO_NODE_IS_DIR(node)) {
+            subdir = (IsoDir *) node;
+            ret = iso_tree_get_node_of_block(image, subdir, block, found, &na,
+                                             1);
+            if (ret != 0)
+                goto ex;
+        }
+    }
+    if (next_above != NULL && (na  > 0 || !(flag & 1)))
+        if (*next_above == 0 || *next_above > na || !(flag & 1))
+            *next_above = na;
+    ret = 0;
+ex:
+    if (sections != NULL)
+        free(sections);
+    if (iter != NULL)
+        iso_dir_iter_free(iter);
+    return ret;
+}
+
+
 /* ------------------------- tree cloning ------------------------------ */
 
 static
@@ -1159,18 +1485,36 @@ int iso_tree_clone_special(IsoSpecial *node,
     return ISO_SUCCESS;
 }
 
-/* API */
-int iso_tree_clone(IsoNode *node,
-                   IsoDir *new_parent, char *new_name, IsoNode **new_node,
-                   int flag)
+
+/* @param flag bit0= Merge directories rather than ISO_NODE_NAME_NOT_UNIQUE.
+               bit1= issue warning in case of truncation
+*/
+int iso_tree_clone_trunc(IsoNode *node, IsoDir *new_parent, 
+                         char *new_name_in, IsoNode **new_node, 
+                         int truncate_length, int flag)
 {
     int ret = ISO_SUCCESS;
+    char *new_name, *trunc = NULL;
 
+    *new_node = NULL;
+    new_name = new_name_in;
+    if (truncate_length >= 64 && (int) strlen(new_name) > truncate_length) {
+        trunc = strdup(new_name);
+        if (trunc == 0) {
+            ret = ISO_OUT_OF_MEM;
+            goto ex;
+        }
+        ret = iso_truncate_rr_name(1, truncate_length, trunc, !(flag & 2));
+        if (ret < 0)
+            goto ex;
+        new_name = trunc;
+    }
     if (iso_dir_get_node(new_parent, new_name, new_node) == 1) {
         if (! (node->type == LIBISO_DIR && (*new_node)->type == LIBISO_DIR &&
                (flag & 1))) {
             *new_node = NULL;
-            return ISO_NODE_NAME_NOT_UNIQUE;
+            ret = ISO_NODE_NAME_NOT_UNIQUE;
+            goto ex;
         }
     } else
         flag &= ~1;
@@ -1191,10 +1535,42 @@ int iso_tree_clone(IsoNode *node,
         ret = ISO_SUCCESS; /* API says they are silently ignored */
     }
     if (ret < 0)
-        return ret;
-    if (flag & 1)
-        return 2; /* merged two directories, *new_node is not new */
+        goto ex;
+    if (flag & 1) {
+        ret = 2; /* merged two directories, *new_node is not new */
+        goto ex;
+    }
     ret = iso_tree_copy_node_attr(node, *new_node, 0);
+
+ex:;
+    if (trunc != NULL)
+        free(trunc);
+    return ret;
+}
+
+
+/* API */
+int iso_tree_clone(IsoNode *node,
+                   IsoDir *new_parent, char *new_name, IsoNode **new_node,
+                   int flag)
+{
+    return iso_tree_clone_trunc(node, new_parent, new_name, new_node, 0,
+                                flag & 1); 
+}
+
+
+/* API */
+int iso_image_tree_clone(IsoImage *image, IsoNode *node, IsoDir *new_parent,
+                         char *new_name, IsoNode **new_node, int flag)
+{
+    int length, ret;
+
+    if (image->truncate_mode == 0)
+        length = 0;
+    else
+        length = image->truncate_length;
+    ret = iso_tree_clone_trunc(node, new_parent, new_name, new_node, length,
+                               flag & 3);
     return ret;
 }
 
