@@ -1,7 +1,7 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
  * Copyright (c) 2007 Mario Danic
- * Copyright (c) 2009 - 2012 Thomas Schmitt
+ * Copyright (c) 2009 - 2015 Thomas Schmitt
  * 
  * This file is part of the libisofs project; you can redistribute it and/or 
  * modify it under the terms of the GNU General Public License version 2 
@@ -17,6 +17,7 @@
 #include "libisofs.h"
 #include "messages.h"
 #include "joliet.h"
+#include "node.h"
 #include "../version.h"
 
 #include <stdlib.h>
@@ -241,8 +242,8 @@ ex:;
     return retval;
 }
 
-int strnconv(const char *str, const char *icharset, const char *ocharset,
-             size_t len, char **output)
+int strnconvl(const char *str, const char *icharset, const char *ocharset,
+              size_t len, char **output, size_t *out_len)
 {
     size_t inbytes;
     size_t outbytes;
@@ -278,7 +279,8 @@ int strnconv(const char *str, const char *icharset, const char *ocharset,
     *ret = '\0';
     iso_iconv_close(&conv, 0);
 
-    *output = malloc(ret - out + 1);
+    *out_len = ret - out;
+    *output = malloc(*out_len + 1);
     if (*output == NULL) {
         retval = ISO_OUT_OF_MEM;
         goto ex;
@@ -291,6 +293,15 @@ ex:;
     return retval;
 }
 
+int strnconv(const char *str, const char *icharset, const char *ocharset,
+             size_t len, char **output)
+{
+    size_t l;
+
+    return strnconvl(str, icharset, ocharset, len, output, &l);
+}
+
+
 /**
  * Convert a str in a specified codeset to WCHAR_T. 
  * The result must be free() when no more needed
@@ -425,6 +436,7 @@ int str2ascii(const char *icharset, const char *input, char **output)
 
         ret_ = malloc(numchars + 1);
         if (ret_ == NULL) {
+            free(wsrc_);
             return ISO_OUT_OF_MEM;
         }
         outbytes = numchars;
@@ -434,7 +446,9 @@ int str2ascii(const char *icharset, const char *input, char **output)
         conv_ret = iso_iconv_open(&conv, "ASCII", "WCHAR_T", 0);
         if (conv_ret <= 0) {
             free(wsrc_);
+            wsrc_ = NULL;
             free(ret_);
+            ret = ret_ = NULL;
         }
     } else if (result != (int) ISO_CHARSET_CONV_ERROR)
         return result;
@@ -583,8 +597,10 @@ int str2ucs(const char *icharset, const char *input, uint16_t **output)
         loop_limit = inbytes + 3;
 
         ret_ = malloc((numchars+1) * sizeof(uint16_t));
-        if (ret_ == NULL)
+        if (ret_ == NULL) {
+            free(wsrc_);
             return ISO_OUT_OF_MEM;
+        }
         outbytes = numchars * sizeof(uint16_t);
         ret = ret_;
 
@@ -592,7 +608,9 @@ int str2ucs(const char *icharset, const char *input, uint16_t **output)
         conv_ret = iso_iconv_open(&conv, "UCS-2BE", "WCHAR_T", 0);
         if (conv_ret <= 0) {
             free(wsrc_);
+            wsrc_ = NULL;
             free(ret_);
+            ret = ret_ = NULL;
         }
     } else if (result != (int) ISO_CHARSET_CONV_ERROR)
         return result;
@@ -713,8 +731,10 @@ int str2utf16be(const char *icharset, const char *input, uint16_t **output)
             loop_limit = inbytes + 3;
 
             ret_ = malloc((2 * numchars+1) * sizeof(uint16_t));
-            if (ret_ == NULL)
+            if (ret_ == NULL) {
+                free(wsrc_);
                 return ISO_OUT_OF_MEM;
+            }
             outbytes = 2 * numchars * sizeof(uint16_t);
             ret = ret_;
 
@@ -1197,7 +1217,7 @@ uint16_t *iso_j_file_id(const uint16_t *src, int flag)
 {
     uint16_t *dot, *retval = NULL;
     size_t lname, lext, lnname, lnext, pos, i, maxchar = 64;
-    uint16_t *dest = NULL;
+    uint16_t *dest = NULL, c;
 
     LIBISO_ALLOC_MEM_VOID(dest, uint16_t, LIBISO_JOLIET_NAME_MAX);
                                /* was: 66 = 64 (name + ext) + 1 (.) + 1 (\0) */
@@ -1237,7 +1257,7 @@ uint16_t *iso_j_file_id(const uint16_t *src, int flag)
 
     /* Convert up to lnname characters of the filename. */
     for (i = 0; i < lnname; i++) {
-        uint16_t c = src[i];
+        c = src[i];
         if (valid_j_char(c)) {
             dest[pos++] = c;
         } else {
@@ -1245,6 +1265,8 @@ uint16_t *iso_j_file_id(const uint16_t *src, int flag)
             pos++;
         }
     }
+    if (pos > 0)
+        iso_handle_split_utf16(dest + (pos - 1));
 
     if ((flag & 1) && lnext <= 0)
         goto is_done;
@@ -1262,6 +1284,7 @@ uint16_t *iso_j_file_id(const uint16_t *src, int flag)
             pos++;
         }
     }
+    iso_handle_split_utf16(dest + (pos - 1));
 
 is_done:;
     set_ucsbe(dest + pos, '\0');
@@ -1298,6 +1321,7 @@ uint16_t *iso_j_dir_id(const uint16_t *src, int flag)
             set_ucsbe(dest + i, '_');
         }
     }
+    iso_handle_split_utf16(dest + (len - 1));
     set_ucsbe(dest + len, '\0');
     retval = ucsdup(dest);
 ex:
@@ -1379,6 +1403,8 @@ uint16_t *ucsncpy(uint16_t *dest, const uint16_t *src, size_t n)
 {
     n = MIN(n, ucslen(src) + 1);
     memcpy(dest, src, n*2);
+    if (n >= 2)
+        iso_handle_split_utf16(dest + (n - 2));
     return dest;
 }
 
@@ -1518,6 +1544,26 @@ uint32_t iso_read_bb(const uint8_t *buf, int bytes, int *error)
     return v1;
 }
 
+uint64_t iso_read_lsb64(const uint8_t *buf)
+{
+    int i;
+    uint64_t ret = 0;
+
+    for (i=0; i < 8; i++)
+        ret += ((uint64_t) buf[i]) << (i * 8);
+    return ret;
+}
+
+uint64_t iso_read_msb64(const uint8_t *buf)
+{
+    int i;
+    uint64_t ret = 0;
+
+    for (i=0; i < 8; i++)
+        ret += ((uint64_t) buf[7 - i]) << (i * 8);
+    return ret;
+}
+
 void iso_datetime_7(unsigned char *buf, time_t t, int always_gmt)
 {
     static int tzsetup = 0;
@@ -1538,8 +1584,14 @@ void iso_datetime_7(unsigned char *buf, time_t t, int always_gmt)
 #else
     if (tm.tm_isdst < 0)
         tm.tm_isdst = 0;
-    tzoffset = ( - timezone / 60 / 15 ) + 4 * tm.tm_isdst;
+#ifndef Libburnia_timezonE
+#define Libburnia_timezonE timezone
 #endif
+ #if Libburnia_timezonE == 0
+    always_gmt = 1;
+ #endif
+    tzoffset = ( - Libburnia_timezonE / 60 / 15 ) + 4 * tm.tm_isdst;
+#endif /* ! HAVE_TM_GMTOFF */
 
     if (tzoffset > 52 || tzoffset < -48 || always_gmt) {
         /* absurd timezone offset, represent time in GMT */
@@ -1584,8 +1636,14 @@ void iso_datetime_17(unsigned char *buf, time_t t, int always_gmt)
 #else
     if (tm.tm_isdst < 0)
         tm.tm_isdst = 0;
-    tzoffset = ( - timezone / 60 / 15 ) + 4 * tm.tm_isdst;
+#ifndef Libburnia_timezonE
+#define Libburnia_timezonE timezone
 #endif
+ #if Libburnia_timezonE == 0
+    always_gmt = 1;
+ #endif
+    tzoffset = ( - Libburnia_timezonE / 60 / 15 ) + 4 * tm.tm_isdst;
+#endif /* ! HAVE_TM_GMTOFF */
 
     if (tzoffset > 52 || tzoffset < -48 || always_gmt) {
         /* absurd timezone offset, represent time in GMT */
@@ -1746,6 +1804,7 @@ time_t iso_datetime_read_7(const uint8_t *buf)
     tm.tm_hour = buf[3];
     tm.tm_min = buf[4];
     tm.tm_sec = buf[5];
+    tm.tm_isdst = 0;
 
     return timegm(&tm) - ((int8_t)buf[6]) * 60 * 15;
 }
@@ -1762,6 +1821,7 @@ time_t iso_datetime_read_17(const uint8_t *buf)
     sscanf((char*)&buf[12], "%2d", &tm.tm_sec);
     tm.tm_year -= 1900;
     tm.tm_mon -= 1;
+    tm.tm_isdst = 0;
 
     return timegm(&tm) - ((int8_t)buf[6]) * 60 * 15;
 }
@@ -1995,6 +2055,17 @@ int iso_util_dec_to_uint32(char *dec, uint32_t *value, int flag)
 }
 
 
+int iso_util_bin_to_hex(char *target, uint8_t *bytes, int num_bytes, int flag)
+{
+    int i;
+
+    for (i = 0; i < num_bytes; i++)
+        sprintf(target + 2 * i, "%-2.2x", bytes[i]);
+    target[2 * num_bytes] = 0;
+    return 1;
+}
+
+
 int iso_util_hex_to_bin(char *hex, char *bin, int bin_size, int *bin_count,
                         int flag)
 {
@@ -2167,6 +2238,7 @@ unexpected_type:;
         goto ex;
     } else if (range_start != ctx_start_lba) {
         ret = ISO_MD5_TAG_MISPLACED;
+        goto ex;
     }
     ret = iso_md5_clone(ctx, &cloned_ctx);
     if (ret < 0)
@@ -2209,3 +2281,170 @@ uint16_t iso_htons(uint16_t v)
     return ret;
 }   
 
+
+/* If an UTF-16 surrogate pair was split : Change to UTF-16 '_'.
+   (UCS-2 is promised to reserve 0xd800 to 0xdbff for UTF-16).
+*/
+void iso_handle_split_utf16(uint16_t *utf_word)
+{
+    unsigned char *hb;
+
+    hb = (unsigned char *) utf_word;
+    if ((hb[0] & 0xfc) == 0xd8)
+        set_ucsbe(utf_word, '_');
+}
+
+
+int iso_clone_mem(char *in, char **out, size_t size)
+{
+    if (in == NULL) {
+        *out = NULL;
+        return 1;
+    }
+    if (size == 0)
+        size = strlen(in) + 1;
+    *out = calloc(1, size);
+    if (*out == NULL)
+        return ISO_OUT_OF_MEM;
+    memcpy(*out, in, size);
+    return ISO_SUCCESS;
+}
+    
+
+int iso_clone_mgtd_mem(char *in, char **out, size_t size)
+{
+    if (*out != NULL)
+        free(*out);
+    return iso_clone_mem(in, out, size);
+}
+
+
+/** Convert a text into a number of type double and multiply it by unit code
+    [kmgt] (2^10 to 2^40) or [s] (2048) or [d] (512).
+    (Also accepts capital letters.)
+    @param text Input like "42", "223062s", "3m" or "-1g"
+    @param flag Bitfield for control purposes:
+                bit0= return -1 rathern than 0 on failure
+                bit1= if scaled then compute the last byte of the last unit
+    @return The derived value
+*/
+off_t iso_scanf_io_size(char *text, int flag)
+{
+    int c;
+    off_t ret = 0, fac = 1;
+    char *rpt;
+
+    for (rpt = text; *rpt >= '0' && *rpt <= '9'; rpt++)
+        ret = ret * 10 + (*rpt - '0');
+    if (rpt == text)
+        return (off_t) (flag & 1 ? -1 : 0);
+    c = *rpt;
+    if (c=='k' || c=='K')
+        fac = 1024;
+    else if (c=='m' || c=='M')
+        fac = 1024 * 1024;
+    else if (c=='g' || c=='G')
+        fac = 1024 * 1024 * 1024;
+    else if (c=='t' || c=='T')
+        fac = ((off_t) 1024) * 1024 * 1024 * 1024;
+    else if (c=='s' || c=='S')
+        fac = 2048;
+    else if (c=='d' || c=='D')
+        fac = 512;
+    ret *= fac;
+    if (flag & 2)
+        ret += fac - 1;
+    return ret;
+}
+
+
+/* Find backward from idx the start byte of a possible UTF-8 character.
+     https://en.wikipedia.org/wiki/UTF-8#Description
+*/
+static
+int find_utf8_start(char *name, int idx, int flag)
+{
+    unsigned char *uname, uch;
+    int i;
+
+    uname= (unsigned char *) name;
+    if ((uname[idx] & 0xc0) != 0x80)
+        return idx;                                /* not an UTF-8 tail byte */
+    for (i = 0; i < 5 && idx - 1 - i >= 0; i++) {                                                                           /* up to deprecated 6-byte codes */
+        uch = uname[idx - 1 - i];
+        if ((uch & 0xe0) == 0xc0 || (uch & 0xf0) == 0xe0 ||
+            (uch & 0xf8) == 0xf0 || (uch & 0xfc) == 0xf8 ||
+            (uch & 0xfe) == 0xfc)
+            return (idx - 1 - i);                  /* UTF-8 start byte found */
+        if ((uch & 0xc0) != 0x80)
+          return idx;                 /* not an UTF-8 tail byte, so no UTF-8 */
+    }
+    return idx;                                      /* no UTF-8 start found */
+}
+
+/* @param flag bit0= do not issue warning message
+*/
+int iso_truncate_rr_name(int truncate_mode, int truncate_length,
+                         char *name, int flag)
+{
+    int neck, goal, ret, l, i;
+    static int hash_size = 32;
+    void *ctx = NULL;
+    char hashval[16];
+
+    l = strlen(name);
+    if (l <= truncate_length)
+        return ISO_SUCCESS;
+    if (truncate_mode == 0)
+        return ISO_RR_NAME_TOO_LONG;
+
+    /* Compute hash */
+    ret = iso_md5_start(&ctx);
+    if (ret < 0)
+        goto ex;
+    ret = iso_md5_compute(ctx, name, l > 4095 ? 4095 : l);
+    if (ret < 0)
+        goto ex;
+    ret = iso_md5_end(&ctx, hashval);
+    if (ret < 0)
+        goto ex;
+
+    if (!(flag & 1))
+        iso_msg_submit(-1, ISO_RR_NAME_TRUNCATED, 0,
+                     "File name had to be truncated and MD5 marked: %s", name);
+
+    /* Avoid to produce incomplete UTF-8 characters */
+    goal = truncate_length - hash_size - 1;
+    neck = find_utf8_start(name, goal, 0);
+    for (; neck < goal; neck++)
+        name[neck] = '_';
+
+    /* Write colon and hash text over end of truncated name */
+    name[goal] = ':';
+    goal++;
+    for (i = 0; goal < truncate_length - 1 && i < hash_size / 2; goal += 2) {
+        sprintf(name + goal, "%2.2x", *((unsigned char *) (hashval + i)));
+        i++;
+    }
+    name[truncate_length] = 0;
+
+    ret = ISO_SUCCESS;
+ex:;
+    if (ctx != NULL)
+        iso_md5_end(&ctx, hashval);
+    return ret;
+}
+
+/* API */
+int iso_truncate_leaf_name(int mode, int length, char *name, int flag)
+{
+    int ret;
+
+    if (mode < 0 || mode > 1)
+        return ISO_WRONG_ARG_VALUE;
+    if (length < 64 || length > LIBISOFS_NODE_NAME_MAX)
+        return ISO_WRONG_ARG_VALUE;
+    ret = iso_truncate_rr_name(mode, length, name, 1);
+    return ret;
+}
+
